# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import copy

import unittest
import collections

from cryptodatahub.common.exception import InvalidValue

from cryptoparser.common.exception import NotEnoughData

from cryptoparser.tls.ldap import (
    LDAPExtendedRequestStartTLS,
    LDAPExtendedResponseStartTLS,
    LDAPResultCode,
)


class TestLDAPExtendedRequest(unittest.TestCase):
    def setUp(self):
        self.ldap_extended_request_dict = collections.OrderedDict([
            ('message_sequence', b'\x30\x1d'),
            ('message_id', b'\x02\x01\x01'),
            ('protocol_op', b'\x77\x18'),
            ('extended_request', b'\x80'),
            ('request_name', (
                b'\x16\x31\x2e\x33\x2e\x36\x2e\x31\x2e\x34\x2e\x31\x2e\x31\x34\x36' +
                b'\x36\x2e\x32\x30\x30\x33\x37'
            )),
        ])
        self.ldap_extended_request_bytes = b''.join(self.ldap_extended_request_dict.values())

        self.ldap_extended_request = LDAPExtendedRequestStartTLS()

    def test_parse(self):
        LDAPExtendedRequestStartTLS.parse_exact_size(self.ldap_extended_request_bytes)

    def test_compose(self):
        self.assertEqual(self.ldap_extended_request.compose(), self.ldap_extended_request_bytes)


class TestLDAPExtendedResponseMinimal(unittest.TestCase):
    def setUp(self):
        self.ldap_extended_response_dict = collections.OrderedDict([
            ('message_sequence', b'\x30\x0c'),
            ('message_id', b'\x02\x01\x01'),
            ('protocol_op', b'\x78\x07'),
            ('extended_response', b''),
            ('result_code', b'\x0a\x01\x07'),
            ('matched_dn', b'\x04\x00'),
            ('diagnostic_message', b'\x04\x00'),
            ('referral', b''),
        ])
        self.ldap_extended_response_bytes = b''.join(self.ldap_extended_response_dict.values())

        self.ldap_extended_response = LDAPExtendedResponseStartTLS(LDAPResultCode.AUTH_METHOD_NOT_SUPPORTED)

    def test_error(self):
        with self.assertRaises(NotEnoughData) as context_manager:
            # pylint: disable=expression-not-assigned
            LDAPExtendedResponseStartTLS.parse_exact_size(
                    self.ldap_extended_response_bytes[:LDAPExtendedResponseStartTLS.HEADER_SIZE]
            )
        self.assertEqual(
            context_manager.exception.bytes_needed,
            len(self.ldap_extended_response_bytes) - LDAPExtendedResponseStartTLS.HEADER_SIZE
        )

        ldap_extended_response_dict = copy.copy(self.ldap_extended_response_dict)
        ldap_extended_response_dict['protocol_op'] = b'\xff\xff'
        ldap_extended_response_bytes = b''.join(ldap_extended_response_dict.values())

        with self.assertRaises(InvalidValue) as context_manager:
            # pylint: disable=expression-not-assigned
            LDAPExtendedResponseStartTLS.parse_exact_size(ldap_extended_response_bytes)

    def test_parse(self):
        ldap_extended_response = LDAPExtendedResponseStartTLS.parse_exact_size(self.ldap_extended_response_bytes)
        self.assertEqual(ldap_extended_response.result_code, self.ldap_extended_response.result_code)

    def test_compose(self):
        self.assertEqual(self.ldap_extended_response.compose(), self.ldap_extended_response_bytes)


class TestLDAPExtendedResponseFull(unittest.TestCase):
    def setUp(self):
        self.ldap_extended_response_dict = collections.OrderedDict([
            ('message_sequence', b'\x30\x34'),
            ('message_id', b'\x02\x01\x01'),
            ('protocol_op', b'\x78\x2f'),
            ('extended_response', b''),
            ('result_code', b'\x0a\x01\x07'),
            ('matched_dn', b'\x04\x08\x00\x01\x02\x03\x04\x05\x06\x07'),
            ('diagnostic_message', b'\x04\x08\x00\x01\x02\x03\x04\x05\x06\x07'),
            ('referral', b''),
            ('response_name', (
                b'\x8a\x16\x31\x2e\x33\x2e\x36\x2e\x31\x2e\x34\x2e\x31\x2e\x31\x34' +
                b'\x36\x36\x2e\x32\x30\x30\x33\x37'
            )),
        ])
        self.ldap_extended_response_bytes = b''.join(self.ldap_extended_response_dict.values())

        self.ldap_extended_response = LDAPExtendedResponseStartTLS(LDAPResultCode.AUTH_METHOD_NOT_SUPPORTED)

    def test_parse(self):
        ldap_extended_response = LDAPExtendedResponseStartTLS.parse_exact_size(self.ldap_extended_response_bytes)
        self.assertEqual(ldap_extended_response.result_code, self.ldap_extended_response.result_code)



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import unittest
import collections

from cryptodatahub.common.exception import InvalidValue
from cryptoparser.common.exception import NotEnoughData, InvalidType

from cryptoparser.tls.rdp import (
    COTPConnectionConfirm,
    COTPConnectionRequest,
    RDPNegotiationRequest,
    RDPNegotiationRequestFlags,
    RDPNegotiationResponse,
    RDPNegotiationResponseFlags,
    RDPProtocol,
    TPKT,
)


class TestTPKT(unittest.TestCase):
    def setUp(self):
        self.tpkt_dict = collections.OrderedDict([
            ('version', b'\x03'),
            ('reserved', b'\x00'),
            ('packet_length', b'\x00\x14'),
            ('message',
             b'\x00\x01\x02\x03\x04\x05\x06\x07' +
             b'\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f' +
             b'')
        ])
        self.tpkt_bytes = b''.join(self.tpkt_dict.values())

        self.tpkt = TPKT(
            version=3,
            message=b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f'
        )

    def test_error(self):
        with self.assertRaises(NotEnoughData) as context_manager:
            TPKT.parse_exact_size(b'')
        self.assertEqual(context_manager.exception.bytes_needed, TPKT.HEADER_SIZE)

        with self.assertRaises(InvalidValue) as context_manager:
            TPKT.parse_exact_size(b'\x01\x00\x00\x00')
        self.assertEqual(context_manager.exception.value, 1)

        with self.assertRaises(NotEnoughData) as context_manager:
            TPKT.parse_exact_size(b'\x03\x00\x00\x05')
        self.assertEqual(context_manager.exception.bytes_needed, 1)

    def test_parse(self):
        tpkt = TPKT.parse_exact_size(self.tpkt_bytes)
        self.assertEqual(tpkt.version, 3)
        self.assertEqual(tpkt.message, self.tpkt_dict['message'])

    def test_compose(self):
        self.assertEqual(self.tpkt.compose(), self.tpkt_bytes)


class TestCOTPConnectionRequest(unittest.TestCase):
    def setUp(self):
        self.cotp_connection_request_dict = collections.OrderedDict([
            ('length', b'\x0e'),
            ('type', b'\xe0'),
            ('src_ref', b'\x01\x02'),
            ('dst_ref', b'\x03\x04'),
            ('class_option', b'\x00'),
            ('user_data', b'\x00\x01\x02\x03\x04\x05\x06\x07'),
        ])
        self.cotp_connection_request_bytes = b''.join(self.cotp_connection_request_dict.values())

        self.cotp_connection_request = COTPConnectionRequest(
            src_ref=0x0102, dst_ref=0x0304, class_option=0x00,
            user_data=b'\x00\x01\x02\x03\x04\x05\x06\x07'
        )

    def test_error(self):
        with self.assertRaises(NotEnoughData) as context_manager:
            COTPConnectionRequest.parse_exact_size(b'')
        self.assertEqual(context_manager.exception.bytes_needed, COTPConnectionRequest.HEADER_SIZE)

        with self.assertRaises(NotEnoughData) as context_manager:
            COTPConnectionRequest.parse_exact_size(b'\x07\xe0\x00\x00\x00\x00\x00')
        self.assertEqual(context_manager.exception.bytes_needed, 1)

        with self.assertRaises(InvalidType) as context_manager:
            COTPConnectionRequest.parse_exact_size(b'\x07\xf0\x00\x00\x00\x00\x00\x00')

        with self.assertRaises(InvalidValue) as context_manager:
            COTPConnectionRequest.parse_exact_size(b'\x07\xe0\x00\x00\x00\x00\x01\x00')
        self.assertEqual(context_manager.exception.value, 1)

    def test_parse(self):
        cotp_connection_request = COTPConnectionRequest.parse_exact_size(self.cotp_connection_request_bytes)
        self.assertEqual(cotp_connection_request.src_ref, self.cotp_connection_request.src_ref)
        self.assertEqual(cotp_connection_request.dst_ref, self.cotp_connection_request.dst_ref)
        self.assertEqual(cotp_connection_request.class_option, self.cotp_connection_request.class_option)
        self.assertEqual(cotp_connection_request.user_data, self.cotp_connection_request.user_data)

    def test_compose(self):
        self.assertEqual(self.cotp_connection_request.compose(), self.cotp_connection_request_bytes)


class TestCOTPConnectionConfirm(unittest.TestCase):
    def setUp(self):
        self.cotp_connection_confirm_dict = collections.OrderedDict([
            ('length', b'\x0e'),
            ('type', b'\xd0'),
            ('src_ref', b'\x01\x02'),
            ('dst_ref', b'\x03\x04'),
            ('class_option', b'\x00'),
            ('user_data', b'\x00\x01\x02\x03\x04\x05\x06\x07'),
        ])
        self.cotp_connection_confirm_bytes = b''.join(self.cotp_connection_confirm_dict.values())

        self.cotp_connection_confirm = COTPConnectionConfirm(
            src_ref=0x0102, dst_ref=0x0304, class_option=0x00,
            user_data=b'\x00\x01\x02\x03\x04\x05\x06\x07'
        )

    def test_parse(self):
        cotp_connection_confirm = COTPConnectionConfirm.parse_exact_size(self.cotp_connection_confirm_bytes)
        self.assertEqual(cotp_connection_confirm.src_ref, self.cotp_connection_confirm.src_ref)
        self.assertEqual(cotp_connection_confirm.dst_ref, self.cotp_connection_confirm.dst_ref)
        self.assertEqual(cotp_connection_confirm.class_option, self.cotp_connection_confirm.class_option)
        self.assertEqual(cotp_connection_confirm.user_data, self.cotp_connection_confirm.user_data)

    def test_compose(self):
        self.assertEqual(self.cotp_connection_confirm.compose(), self.cotp_connection_confirm_bytes)


class TestRDPNegotiationRequest(unittest.TestCase):
    def setUp(self):
        self.rdp_negotiation_request_dict = collections.OrderedDict([
            ('type', b'\x01'),
            ('flags', b'\x03'),
            ('length', b'\x08\x00'),
            ('protocol', b'\x03\x00\x00\x00'),
        ])
        self.rdp_negotiation_request_bytes = b''.join(self.rdp_negotiation_request_dict.values())

        self.rdp_negotiation_request = RDPNegotiationRequest(
            flags={
                RDPNegotiationRequestFlags.RESTRICTED_ADMIN_MODE_REQUIRED,
                RDPNegotiationRequestFlags.REDIRECTED_AUTHENTICATION_MODE_REQUIRED
            },
            protocol={RDPProtocol.SSL, RDPProtocol.HYBRID}
        )

    def test_error(self):
        with self.assertRaises(NotEnoughData) as context_manager:
            RDPNegotiationRequest.parse_exact_size(b'\x01')
        self.assertEqual(context_manager.exception.bytes_needed, RDPNegotiationRequest.PACKET_LENGTH - 1)

        with self.assertRaises(InvalidType) as context_manager:
            RDPNegotiationRequest.parse_exact_size(b'\x02\x03\x08\x00\x00\x00\x00\x00')

        with self.assertRaises(NotEnoughData) as context_manager:
            RDPNegotiationRequest.parse_exact_size(b'\x01\x03\x08\x00\x00\x00')
        self.assertEqual(context_manager.exception.bytes_needed, 2)

        with self.assertRaises(InvalidValue) as context_manager:
            RDPNegotiationRequest.parse_exact_size(b'\x01\x03\x02\x00\x00\x00\x00\x00')
        self.assertEqual(context_manager.exception.value, 2)

    def test_parse(self):
        rdp_negotiation_request = RDPNegotiationRequest.parse_exact_size(self.rdp_negotiation_request_bytes)
        self.assertEqual(rdp_negotiation_request.flags, self.rdp_negotiation_request.flags)
        self.assertEqual(rdp_negotiation_request.protocol, self.rdp_negotiation_request.protocol)

    def test_compose(self):
        self.assertEqual(self.rdp_negotiation_request.compose(), self.rdp_negotiation_request_bytes)


class TestRDPNegotiationResponse(unittest.TestCase):
    def setUp(self):
        self.rdp_negotiation_request_dict = collections.OrderedDict([
            ('type', b'\x02'),
            ('flags', b'\x03'),
            ('length', b'\x08\x00'),
            ('protocol', b'\x03\x00\x00\x00'),
        ])
        self.rdp_negotiation_request_bytes = b''.join(self.rdp_negotiation_request_dict.values())

        self.rdp_negotiation_request = RDPNegotiationResponse(
            flags={
                RDPNegotiationResponseFlags.EXTENDED_CLIENT_DATA_SUPPORTED,
                RDPNegotiationResponseFlags.DYNVC_GFX_PROTOCOL_SUPPORTED
            },
            protocol={RDPProtocol.SSL, RDPProtocol.HYBRID}
        )

    def test_parse(self):
        rdp_negotiation_request = RDPNegotiationResponse.parse_exact_size(self.rdp_negotiation_request_bytes)
        self.assertEqual(rdp_negotiation_request.flags, self.rdp_negotiation_request.flags)
        self.assertEqual(rdp_negotiation_request.protocol, self.rdp_negotiation_request.protocol)

    def test_compose(self):
        self.assertEqual(self.rdp_negotiation_request.compose(), self.rdp_negotiation_request_bytes)



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import abc
import collections
import datetime
import enum

import attr

from cryptodatahub.common.algorithm import Authentication, NamedGroup, Signature
from cryptodatahub.common.exception import InvalidValue
from cryptodatahub.common.key import (
    PublicKey,
    PublicKeyParamsDsa,
    PublicKeyParamsEcdsa,
    PublicKeyParamsEddsa,
    PublicKeyParamsRsa,
)

from cryptodatahub.dnsrec.algorithm import DnsRrType, DnsSecAlgorithm, DnsSecDigestType

from cryptoparser.common.base import NumericRangeParsableBase, OneByteEnumParsable, Serializable, TwoByteEnumParsable
from cryptoparser.common.exception import NotEnoughData
from cryptoparser.common.parse import ByteOrder, ComposerBinary, ParsableBase, ParserBinary


class DnsSecProtocol(enum.Enum):
    V3 = 3


class DnsSecFlag(enum.IntEnum):
    SECURE_ENTRY_POINT = 0x0001
    REVOKE = 0x0080
    DNS_ZONE_KEY = 0x0100


class DnsSecAlgorithmFactory(OneByteEnumParsable):
    @classmethod
    def get_enum_class(cls):
        return DnsSecAlgorithm

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()


@attr.s
class DnsRecordDnskey(ParsableBase, Serializable):
    HEADER_SIZE = 4

    flags = attr.ib(validator=attr.validators.deep_iterable(attr.validators.instance_of(DnsSecFlag)))
    algorithm = attr.ib(validator=attr.validators.instance_of(DnsSecAlgorithm))
    key = attr.ib(validator=attr.validators.instance_of(PublicKey))
    protocol = attr.ib(validator=attr.validators.instance_of(DnsSecProtocol))

    def __attrs_post_init__(self):
        if not isinstance(self.algorithm.value.algorithm, Signature):
            raise InvalidValue(self.algorithm.value.algorithm, type(self), 'algorithm_type')

        algorithm_key_type = self.algorithm.value.algorithm.value.key_type
        algorithm_incompatible = self.key.key_type != algorithm_key_type
        algorithm_compatible_gost = (
            self.algorithm == DnsSecAlgorithm.ECCGOST and
            self.key.key_type == Authentication.ECDSA and
            self.key.params.named_group == NamedGroup.GC256B
        )
        if (algorithm_incompatible and not algorithm_compatible_gost):
            raise InvalidValue(algorithm_key_type, type(self), 'key_type')

    @property
    def key_tag(self):
        if self.algorithm == DnsSecAlgorithm.RSAMD5:
            return (self.key.params.modulus & 0xffffff) >> 8

        key_tag = 0
        parser = ParserBinary(self.compose(), byte_order=ByteOrder.BIG_ENDIAN)

        while parser.unparsed_length > 1:
            parser.parse_numeric('value', 2)
            key_tag += parser['value']
        if parser.unparsed_length:
            parser.parse_numeric('value', 1)
            key_tag += parser['value']

        key_tag += (key_tag >> 16) & 0xffff
        return key_tag & 0xffff

    def _asdict(self):
        dict_value = super()._asdict()
        return collections.OrderedDict([('key_tag', self.key_tag)] + list(dict_value.items()))

    @classmethod
    def _parse_public_key_rsa(cls, key_parser):
        key_parser.parse_numeric('exponent_length_one_octet', 1)
        exponent_length = key_parser['exponent_length_one_octet']
        if exponent_length == 0:
            key_parser.parse_numeric('exponent_length_two_octets', 2)
            exponent_length = key_parser['exponent_length_two_octets']
        key_parser.parse_mpint('public_exponent', exponent_length)
        key_parser.parse_mpint('modulus', key_parser.unparsed_length)

        return PublicKey.from_params(PublicKeyParamsRsa(
            public_exponent=key_parser['public_exponent'],
            modulus=key_parser['modulus'],
        ))

    @classmethod
    def _parse_public_key_ecdsa(cls, dnssec_algorithm, key_parser):
        if dnssec_algorithm == DnsSecAlgorithm.ECDSAP256SHA256:
            named_group = NamedGroup.SECP256K1
        elif dnssec_algorithm == DnsSecAlgorithm.ECDSAP384SHA384:
            named_group = NamedGroup.SECP384R1
        elif dnssec_algorithm == DnsSecAlgorithm.ECCGOST:
            named_group = NamedGroup.GC256B
        else:
            raise NotImplementedError(dnssec_algorithm)

        key_size = named_group.value.size // 8
        key_parser.parse_mpint('x', key_size)
        key_parser.parse_mpint('y', key_size)

        return PublicKey.from_params(PublicKeyParamsEcdsa(
            point_x=key_parser['x'], point_y=key_parser['y'], named_group=named_group,
        ))

    @classmethod
    def _parse_public_key_eddsa(cls, dnssec_algorithm, key_parser):
        if dnssec_algorithm == DnsSecAlgorithm.ED25519:
            curve_type = NamedGroup.CURVE25519
            key_parser.parse_raw('public_key', 256 // 8)
        elif dnssec_algorithm == DnsSecAlgorithm.ED448:
            curve_type = NamedGroup.CURVE448
            key_parser.parse_raw('public_key', 448 // 8)
        else:
            raise NotImplementedError(dnssec_algorithm)

        return PublicKey.from_params(PublicKeyParamsEddsa(
            curve_type=curve_type, key_data=key_parser['public_key']
        ))

    @classmethod
    def _parse_public_key_dss(cls, key_parser):
        key_parser.parse_numeric('t', 1)
        key_parser.parse_mpint('q', 20)

        mpint_length = 64 + key_parser['t'] * 8
        key_parser.parse_mpint('p', mpint_length)
        key_parser.parse_mpint('g', mpint_length)
        key_parser.parse_mpint('y', mpint_length)

        return PublicKey.from_params(PublicKeyParamsDsa(
            prime=key_parser['p'],
            generator=key_parser['g'],
            order=key_parser['q'],
            public_key_value=key_parser['y'],
        ))

    @classmethod
    def parse_key(cls, parsable, dnssec_algorithm):
        key_parser = ParserBinary(parsable)

        public_key_type = dnssec_algorithm.value.algorithm.value.key_type
        if public_key_type == Authentication.RSA:
            public_key = cls._parse_public_key_rsa(key_parser)
        elif public_key_type in [Authentication.ECDSA, Authentication.GOST_R3410_01]:
            public_key = cls._parse_public_key_ecdsa(dnssec_algorithm, key_parser)
        elif public_key_type == Authentication.EDDSA:
            public_key = cls._parse_public_key_eddsa(dnssec_algorithm, key_parser)
        elif public_key_type == Authentication.DSS:
            public_key = cls._parse_public_key_dss(key_parser)
        else:
            raise NotImplementedError(dnssec_algorithm)

        return public_key

    @classmethod
    def _parse(cls, parsable):
        if len(parsable) < cls.HEADER_SIZE:
            raise NotEnoughData(cls.HEADER_SIZE - len(parsable))

        parser = ParserBinary(parsable)

        parser.parse_numeric_flags('flags', 2, DnsSecFlag)
        parser.parse_numeric('protocol', 1, DnsSecProtocol)

        parser.parse_parsable('algorithm', DnsSecAlgorithmFactory)
        parser.parse_raw('key', parser.unparsed_length)

        public_key = cls.parse_key(parser['key'], parser['algorithm'])

        return cls(
            parser['flags'],
            parser['algorithm'],
            public_key,
            parser['protocol'],
        ), parser.parsed_length

    @staticmethod
    def _compose_public_key_rsa(key_composer, key):
        key_params = key.params
        exponent_length = (key_params.public_exponent.bit_length() + 7) // 8
        if exponent_length > 255:
            key_composer.compose_numeric(0, 1)
            key_composer.compose_numeric(exponent_length, 2)
        else:
            key_composer.compose_numeric(exponent_length, 1)

        key_composer.compose_mpint(key_params.public_exponent, exponent_length)
        key_composer.compose_mpint(key_params.modulus, key.key_size // 8)

    @staticmethod
    def _compose_public_key_ecdsa(key_composer, key):
        key_params = key.params
        key_size = key.key_size // 8
        key_composer.compose_mpint(key_params.point_x, key_size)
        key_composer.compose_mpint(key_params.point_y, key_size)

    @staticmethod
    def _compose_public_key_eddsa(key_composer, key):
        key_params = key.params
        key_composer.compose_raw(key_params.key_data)

    @staticmethod
    def _compose_public_key_dss(key_composer, key):
        key_params = key.params
        key_size = key.key_size // 8

        key_composer.compose_numeric((key_size - 64) // 8, 1)
        key_composer.compose_mpint(key_params.order, 20)

        key_composer.compose_mpint(key_params.prime, key_size)
        key_composer.compose_mpint(key_params.generator, key_size)
        key_composer.compose_mpint(key_params.public_key_value, key_size)

    @staticmethod
    def compose_key(key):
        key_composer = ComposerBinary()
        public_key_type = key.key_type

        if public_key_type == Authentication.RSA:
            DnsRecordDnskey._compose_public_key_rsa(key_composer, key)
        elif public_key_type in [Authentication.ECDSA, Authentication.GOST_R3410_01]:
            DnsRecordDnskey._compose_public_key_ecdsa(key_composer, key)
        elif public_key_type == Authentication.EDDSA:
            DnsRecordDnskey._compose_public_key_eddsa(key_composer, key)
        elif public_key_type == Authentication.DSS:
            DnsRecordDnskey._compose_public_key_dss(key_composer, key)
        else:
            raise NotImplementedError(public_key_type)

        return key_composer.composed_bytes

    def compose(self):
        composer = ComposerBinary()

        composer.compose_numeric_flags(self.flags, 2)
        composer.compose_numeric(self.protocol.value, 1)

        composer.compose_numeric_enum_coded(self.algorithm)

        key_bytes = self.compose_key(self.key)

        return composer.composed_bytes + key_bytes


class DnsSecDigestTypeFactory(OneByteEnumParsable):
    @classmethod
    def get_enum_class(cls):
        return DnsSecDigestType

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()


@attr.s
class DnsRecordDs(ParsableBase):
    HEADER_SIZE = 4

    key_tag = attr.ib(validator=attr.validators.instance_of(int))
    algorithm = attr.ib(validator=attr.validators.instance_of(DnsSecAlgorithm))
    digest_type = attr.ib(validator=attr.validators.instance_of(DnsSecDigestType))
    digest = attr.ib(validator=attr.validators.instance_of((bytes, bytearray)))

    @classmethod
    def _parse(cls, parsable):
        if len(parsable) < cls.HEADER_SIZE:
            raise NotEnoughData(cls.HEADER_SIZE - len(parsable))

        parser = ParserBinary(parsable)

        parser.parse_numeric('key_tag', 2)
        parser.parse_parsable('algorithm', DnsSecAlgorithmFactory)
        parser.parse_parsable('digest_type', DnsSecDigestTypeFactory)
        parser.parse_raw('digest', parser.unparsed_length)

        return cls(**parser), parser.parsed_length

    def compose(self):
        composer = ComposerBinary()

        composer.compose_numeric(self.key_tag, 2)
        composer.compose_numeric_enum_coded(self.algorithm)
        composer.compose_numeric_enum_coded(self.digest_type)
        composer.compose_raw(self.digest)

        return composer.composed_bytes


class DnsRrTypeFactory(TwoByteEnumParsable):
    @classmethod
    def get_enum_class(cls):
        return DnsRrType

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()


class DnsRrTypePrivate(NumericRangeParsableBase):
    @classmethod
    def _get_value_min(cls):
        return 0xff00

    @classmethod
    def _get_value_max(cls):
        return 0xfffe

    @classmethod
    def _get_value_length(cls):
        return 2


@attr.s
class DnsNameUncompressed(ParsableBase, Serializable):
    labels = attr.ib(
        validator=attr.validators.deep_iterable(member_validator=attr.validators.instance_of(str))
    )

    def __str__(self):
        return '.'.join(self.labels)

    def _as_markdown(self, level):
        return self._markdown_result(str(self), level)

    @classmethod
    def convert(cls, value):
        if isinstance(value, cls):
            return value
        if isinstance(value, str):
            if not value:
                return cls([])

            return cls(value.split('.'))

        raise InvalidValue(value, cls, 'labels')

    @classmethod
    def _parse(cls, parsable):
        parser = ParserBinary(parsable)

        labels = []
        while True:
            parser.parse_string('label', 1, encoding='idna')
            label = parser['label']

            if not label:
                break

            labels.append(label)

        return cls(labels), parser.parsed_length

    def compose(self):
        composer = ComposerBinary()

        for label in self.labels:
            composer.compose_string(label, 'idna', 1)

        composer.compose_numeric(0, 1)

        return composer.composed_bytes


@attr.s
class DnsRecordRrsig(ParsableBase):  # pylint: disable=too-many-instance-attributes
    HEADER_SIZE = 24

    type_covered = attr.ib(validator=attr.validators.instance_of((DnsRrType, DnsRrTypePrivate)))
    algorithm = attr.ib(validator=attr.validators.instance_of(DnsSecAlgorithm))
    labels = attr.ib(validator=attr.validators.instance_of(int))
    original_ttl = attr.ib(
        validator=attr.validators.instance_of(int),
        metadata={'human_readable_name': 'Original TTL'}
    )
    signature_expiration = attr.ib(validator=attr.validators.instance_of(datetime.datetime))
    signature_inception = attr.ib(validator=attr.validators.instance_of(datetime.datetime))
    key_tag = attr.ib(validator=attr.validators.instance_of(int))
    signers_name = attr.ib(
        converter=DnsNameUncompressed.convert,
        validator=attr.validators.instance_of(DnsNameUncompressed)
    )
    signature = attr.ib(
        validator=attr.validators.instance_of((bytes, bytearray)),
        metadata={'human_friendly': False}
    )

    @classmethod
    def _parse(cls, parsable):
        if len(parsable) < cls.HEADER_SIZE:
            raise NotEnoughData(cls.HEADER_SIZE - len(parsable))

        parser = ParserBinary(parsable)

        try:
            parser.parse_parsable('type_covered', DnsRrTypeFactory)
        except InvalidValue:
            parser.parse_parsable('type_covered', DnsRrTypePrivate)
        parser.parse_parsable('algorithm', DnsSecAlgorithmFactory)
        parser.parse_numeric('labels', 1)
        parser.parse_numeric('original_ttl', 4)
        parser.parse_timestamp('signature_expiration', item_size=4)
        parser.parse_timestamp('signature_inception', item_size=4)
        parser.parse_numeric('key_tag', 2)
        parser.parse_parsable('signers_name', DnsNameUncompressed)
        parser.parse_raw('signature', parser.unparsed_length)

        return cls(**parser), parser.parsed_length

    def compose(self):
        composer = ComposerBinary()

        if isinstance(self.type_covered, DnsRrType):
            composer.compose_numeric_enum_coded(self.type_covered)
        else:
            composer.compose_parsable(self.type_covered)
        composer.compose_numeric_enum_coded(self.algorithm)
        composer.compose_numeric(self.labels, 1)
        composer.compose_numeric(self.original_ttl, 4)
        composer.compose_timestamp(self.signature_expiration, item_size=4)
        composer.compose_timestamp(self.signature_inception, item_size=4)
        composer.compose_numeric(self.key_tag, 2)
        composer.compose_parsable(self.signers_name)
        composer.compose_raw(self.signature)

        return composer.composed_bytes


@attr.s
class DnsRecordMx(ParsableBase):
    HEADER_SIZE = 2

    priority = attr.ib(validator=attr.validators.instance_of(int))
    exchange = attr.ib(
        converter=DnsNameUncompressed.convert,
        validator=attr.validators.instance_of(DnsNameUncompressed)
    )

    @classmethod
    def _parse(cls, parsable):
        if len(parsable) < cls.HEADER_SIZE:
            raise NotEnoughData(cls.HEADER_SIZE - len(parsable))

        parser = ParserBinary(parsable)

        parser.parse_numeric('priority', 2)
        parser.parse_parsable('exchange', DnsNameUncompressed)

        return cls(**parser), parser.parsed_length

    def compose(self):
        composer = ComposerBinary()

        composer.compose_numeric(self.priority, 2)
        composer.compose_parsable(self.exchange)

        return composer.composed_bytes


@attr.s
class DnsRecordTxt(ParsableBase):
    HEADER_SIZE = 1

    value = attr.ib(validator=attr.validators.instance_of(str))

    @classmethod
    def _parse(cls, parsable):
        if len(parsable) < cls.HEADER_SIZE:
            raise NotEnoughData(cls.HEADER_SIZE - len(parsable))

        parser = ParserBinary(parsable)
        value = ''

        while parser.unparsed_length:
            parser.parse_string('value', 1, encoding='ascii')
            value += parser['value']

        return cls(value), parser.parsed_length

    def compose(self):
        composer = ComposerBinary()

        composer.compose_string(self.value, 'ascii', 1)

        return composer.composed_bytes



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import abc
import enum
import re

import attr

import asn1crypto.core

from cryptodatahub.common.exception import InvalidValue

from cryptoparser.common.exception import NotEnoughData
from cryptoparser.common.parse import ParsableBase


class LDAPClass(enum.IntEnum):
    UNIVERSAL = 0
    APPLICATION = 1
    CONTEXT = 2


class LDAPResultCode(enum.IntEnum):
    SUCCESS = 0
    OPERATIONS_ERROR = 1
    PROTOCOL_ERROR = 2
    TIME_LIMIT_EXCEEDED = 3
    SIZE_LIMIT_EXCEEDED = 4
    COMPARE_FALSE = 5
    COMPARE_TRUE = 6
    AUTH_METHOD_NOT_SUPPORTED = 7
    STRONGER_AUTH_REQUIRED = 8
    REFERRAL = 10
    ADMIN_LIMIT_EXCEEDED = 11
    UNAVAILABLE_CRITICAL_EXTENSION = 12
    CONFIDENTIALITY_REQUIRED = 13
    SASL_BIND_IN_PROGRESS = 14
    NO_SUCH_ATTRIBUTE = 16
    UNDEFINED_ATTRIBUTE_TYPE = 17
    INAPPROPRIATE_MATCHING = 18
    CONSTRAINT_VIOLATION = 19
    ATTRIBUTE_OR_VALUE_EXISTS = 20
    INVALID_ATTRIBUTE_SYNTAX = 21
    NO_SUCH_OBJECT = 32
    ALIAS_PROBLEM = 33
    INVALID_DN_SYNTAX = 34
    ALIAS_DEREFERENCING_PROBLEM = 36
    INAPPROPRIATE_AUTHENTICATION = 48
    INVALID_CREDENTIALS = 49
    INSUFFICIENT_ACCESS_RIGHTS = 50
    BUSY = 51
    UNAVAILABLE = 52
    UNWILLING_TO_PERFORM = 53
    LOOP_DETECT = 54
    NAMING_VIOLATION = 64
    OBJECT_CLASS_VIOLATION = 65
    NOT_ALLOWED_ON_NON_LEAF = 66
    NOT_ALLOWED_ON_RDN = 67
    ENTRY_ALREADY_EXISTS = 68
    OBJECT_CLASS_MODS_PROHIBITED = 69
    AFFECTS_MULTIPLE_DSAS = 71
    OTHER = 80


class LDAPResultCodeEnum(asn1crypto.core.Enumerated):
    _map = {result_code.value: result_code for result_code in list(LDAPResultCode)}


class LDAPOID(asn1crypto.core.OctetString):
    pass


class LDAPControl(asn1crypto.core.Sequence):
    _fields = [
        ('controlType', LDAPOID),
        ('criticality', asn1crypto.core.Boolean, {'default': False}),
        ('controlValue', asn1crypto.core.OctetString, {'optional': True}),
    ]


class LDAPControls(asn1crypto.core.SequenceOf):
    _child_spec = LDAPControl


class LDAPExtendedRequest(asn1crypto.core.Sequence):
    _fields = [
        ('requestName', LDAPOID, {'implicit': (LDAPClass.CONTEXT.value, 0)}),
        ('requestValue', asn1crypto.core.OctetString, {'implicit': (LDAPClass.CONTEXT.value, 1), 'optional': True}),
    ]


class LDAPDN(asn1crypto.core.OctetString):
    pass


class LDAPString(asn1crypto.core.OctetString):
    pass


class LDAPURI(LDAPString):
    pass


class LDAPReferral(asn1crypto.core.SequenceOf):
    _child_spec = LDAPURI


class LDAPExtendedResponse(asn1crypto.core.Sequence):
    _fields = [
        ('resultCode', LDAPResultCodeEnum),
        ('matchedDN', LDAPDN),
        ('diagnosticMessage', LDAPString),
        ('referral', LDAPReferral,  {'implicit': (LDAPClass.CONTEXT.value, 3), 'optional': True}),
        ('responseName', LDAPOID, {'implicit': (LDAPClass.CONTEXT.value, 10), 'optional': True}),
        ('responseValue', asn1crypto.core.OctetString, {'implicit': (LDAPClass.CONTEXT.value, 11), 'optional': True}),
    ]


class LDAPProtocolOp(asn1crypto.core.Choice):
    _alternatives = [
        ('extendedReq', LDAPExtendedRequest, {'implicit': (LDAPClass.APPLICATION.value, 23)}),
        ('extendedResp', LDAPExtendedResponse, {'implicit': (LDAPClass.APPLICATION.value, 24)}),
    ]


class LDAPMessage(asn1crypto.core.Sequence):
    _fields = [
        ('messageID', asn1crypto.core.Integer),
        ('protocolOp', LDAPProtocolOp),
        ('controls', LDAPControls, {'implicit': (LDAPClass.CONTEXT.value, 0), 'optional': True}),
    ]


class LDAPMessageParsableBase(ParsableBase):
    HEADER_SIZE = 6

    _NOT_ENOUGH_DATA_REGEX = re.compile(
        r'Insufficient data - ([0-9]+) bytes requested but only ([0-9]+) available'
    )

    @classmethod
    @abc.abstractmethod
    def _parse(cls, parsable):
        raise NotImplementedError()

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()

    @classmethod
    def _parse_asn1(cls, parsable):
        try:
            message = LDAPMessage.load(bytes(parsable))
            # ensure recursive parsing
            message.native  # pylint: disable=pointless-statement
        except ValueError as e:
            match = cls._NOT_ENOUGH_DATA_REGEX.match(e.args[0])
            if match:
                bytes_requested = int(match.group(1))
                bytes_available = int(match.group(2))
                raise NotEnoughData(bytes_requested - bytes_available) from e

            raise InvalidValue(parsable, cls) from e

        return message


class LDAPExtendedRequestStartTLS(LDAPMessageParsableBase):
    @classmethod
    def _parse(cls, parsable):
        asn1_message = cls._parse_asn1(parsable)

        return LDAPExtendedRequestStartTLS(), len(asn1_message.dump())

    def compose(self):
        return LDAPMessage({
            'messageID': 1,
            'protocolOp': {
                'extendedReq': {
                    'requestName': b'1.3.6.1.4.1.1466.20037'
                }
            }
        }).dump()


@attr.s
class LDAPExtendedResponseStartTLS(LDAPMessageParsableBase):
    result_code = attr.ib(validator=attr.validators.in_(LDAPResultCode))

    @classmethod
    def _parse(cls, parsable):
        asn1_message = cls._parse_asn1(parsable)

        return LDAPExtendedResponseStartTLS(
            asn1_message['protocolOp'].chosen['resultCode'].native
        ), len(asn1_message.dump())

    def compose(self):
        return LDAPMessage({
            'messageID': 1,
            'protocolOp': {
                'extendedResp': {
                    'resultCode': self.result_code.value,
                    'matchedDN': b'',
                    'diagnosticMessage': b''
                }
            }
        }).dump()



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import collections
import unittest


from cryptodatahub.common.exception import InvalidValue
from cryptodatahub.common.key import PublicKey, PublicKeyParamsRsa
from cryptodatahub.ssh.algorithm import (
    SshCompressionAlgorithm,
    SshEncryptionAlgorithm,
    SshHostKeyAlgorithm,
    SshKexAlgorithm,
    SshMacAlgorithm,
)

from cryptoparser.common.classes import LanguageTag
from cryptoparser.common.exception import TooMuchData

from cryptoparser.ssh.key import SshHostKeyRSA
from cryptoparser.ssh.subprotocol import (
    SshDHGroupExchangeInit,
    SshDHGroupExchangeGroup,
    SshDHGroupExchangeReply,
    SshDHGroupExchangeRequest,
    SshDHKeyExchangeInit,
    SshDHKeyExchangeReply,
    SshKeyExchangeInit,
    SshMessageVariantInit,
    SshMessageVariantKexDH,
    SshMessageVariantKexDHGroup,
    SshNewKeys,
    SshProtocolMessage,
    SshUnimplementedMessage,
)
from cryptoparser.ssh.version import SshProtocolVersion, SshSoftwareVersionUnparsed, SshVersion


class TestProtocolMessage(unittest.TestCase):
    def test_error(self):
        with self.assertRaises(InvalidValue) as context_manager:
            SshProtocolMessage.parse_exact_size(b'ABC')
        self.assertEqual(context_manager.exception.value, 'ABC')

        with self.assertRaises(InvalidValue) as context_manager:
            SshProtocolMessage.parse_exact_size(b'SSH-2.0\r\n')
        self.assertEqual(context_manager.exception.value, b'\r')

        with self.assertRaises(InvalidValue) as context_manager:
            SshProtocolMessage.parse_exact_size(b'SSH-2.0-software_version\r')
        self.assertEqual(context_manager.exception.value, b'software_version\r')

        with self.assertRaises(TooMuchData) as context_manager:
            SshProtocolMessage.parse_exact_size(b'SSH-2.0-software_version ' + b'X' * 255 + b'\r\n')
        self.assertEqual(context_manager.exception.bytes_needed, len(b'SSH-2.0-software_version ') + len(b'\r\n'))

    def test_parse(self):
        message = SshProtocolMessage.parse_exact_size(b'SSH-1.1-software_version\r\n')
        self.assertEqual(message.protocol_version, SshProtocolVersion(SshVersion.SSH1, 1))
        self.assertEqual(message.software_version, SshSoftwareVersionUnparsed('software_version'))
        self.assertEqual(message.comment, None)

        message = SshProtocolMessage.parse_exact_size(b'SSH-2.2-software_version\r\n')
        self.assertEqual(message.protocol_version, SshProtocolVersion(SshVersion.SSH2, 2))
        self.assertEqual(message.software_version, SshSoftwareVersionUnparsed('software_version'))
        self.assertEqual(message.comment, None)

        message = SshProtocolMessage.parse_exact_size(b'SSH-2.0-software_version comment\r\n')
        self.assertEqual(message.protocol_version, SshProtocolVersion(SshVersion.SSH2, 0))
        self.assertEqual(message.software_version, SshSoftwareVersionUnparsed('software_version'))
        self.assertEqual(message.comment, 'comment')

        message = SshProtocolMessage.parse_exact_size(b'SSH-2.0-software_version comment with spaces\r\n')
        self.assertEqual(message.protocol_version, SshProtocolVersion(SshVersion.SSH2, 0))
        self.assertEqual(message.software_version, SshSoftwareVersionUnparsed('software_version'))
        self.assertEqual(message.comment, 'comment with spaces')

    def test_comment(self):
        with self.assertRaises(InvalidValue):
            SshProtocolMessage(
                SshProtocolVersion(SshVersion.SSH2, 2),
                SshSoftwareVersionUnparsed('software_version'),
                comment='αβγ',
            )
        with self.assertRaises(InvalidValue):
            SshProtocolMessage(
                SshProtocolVersion(SshVersion.SSH2, 2),
                SshSoftwareVersionUnparsed('software_version'),
                comment='comment\r',
            )
        with self.assertRaises(InvalidValue):
            SshProtocolMessage(
                SshProtocolVersion(SshVersion.SSH2, 2),
                SshSoftwareVersionUnparsed('software_version'),
                comment='comment\n',
            )

    def test_compose(self):
        self.assertEqual(
            SshProtocolMessage(
                SshProtocolVersion(SshVersion.SSH2, 2),
                SshSoftwareVersionUnparsed('software_version'),
            ).compose(),
            b'SSH-2.2-software_version\r\n'
        )

        self.assertEqual(
            SshProtocolMessage(
                SshProtocolVersion(SshVersion.SSH2, 2),
                SshSoftwareVersionUnparsed('software_version'),
                'comment'
            ).compose(),
            b'SSH-2.2-software_version comment\r\n'
        )

        self.assertEqual(
            SshProtocolMessage(
                SshProtocolVersion(SshVersion.SSH2, 2),
                SshSoftwareVersionUnparsed('software_version'),
                'comment with spaces'
            ).compose(),
            b'SSH-2.2-software_version comment with spaces\r\n'
        )


class TestKeyExchangeInitMessage(unittest.TestCase):
    def setUp(self):
        self.key_exchange_init_bytes = bytes(
            b'\x14' +                                               # message_code = SshMessageCode.KEXINIT
            b'\x00\x01\x02\x03\x04\x05\x06\x07' +
            b'\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f' +                   # cookie
            b'\x00\x00\x00\x3d' +
            b'diffie-hellman-group1-sha1,ecdh-sha2-nistp256,unparsable-algo' +  # kex_algorithms
            b'\x00\x00\x00\x2f' +
            b'ssh-ed25519,ecdsa-sha2-nistp256,unparsable-algo' +    # host_key_algorithms
            b'\x00\x00\x00\x31' +
            b'aes128-cbc,aes256-gcm@openssh.com,unparsable-algo' +  # encryption_algorithms_client_to_server
            b'\x00\x00\x00\x31' +
            b'aes256-gcm@openssh.com,aes128-cbc,unparsable-algo' +  # encryption_algorithms_server_to_client
            b'\x00\x00\x00\x2e' +
            b'hmac-sha1,umac-128@openssh.com,unparsable-algo' +     # mac_algorithms_client_to_server
            b'\x00\x00\x00\x2e' +
            b'umac-128@openssh.com,hmac-sha1,unparsable-algo' +     # mac_algorithms_server_to_client
            b'\x00\x00\x00\x25' +
            b'none,zlib@openssh.com,unparsable-algo' +              # compression_algorithms_client_to_server
            b'\x00\x00\x00\x25' +
            b'zlib@openssh.com,none,unparsable-algo' +              # compression_algorithms_server_to_client
            b'\x00\x00\x00\x0b' +
            b'en-UK,en-US' +                                        # languages_client_to_server
            b'\x00\x00\x00\x0b' +
            b'en-US,en-UK' +                                        # languages_server_to_client
            b'\x00' +                                               # first_kex_packet_follows
            b'\x00\x01\x02\x03' +                                   # reserved
            b''
        )
        self.key_exchange_init = SshKeyExchangeInit(
            kex_algorithms=[
                SshKexAlgorithm.DIFFIE_HELLMAN_GROUP1_SHA1,
                SshKexAlgorithm.ECDH_SHA2_NISTP256,
                'unparsable-algo',
            ],
            host_key_algorithms=[
                SshHostKeyAlgorithm.SSH_ED25519,
                SshHostKeyAlgorithm.ECDSA_SHA2_NISTP256,
                'unparsable-algo',
            ],
            encryption_algorithms_client_to_server=[
                SshEncryptionAlgorithm.AES128_CBC,
                SshEncryptionAlgorithm.AES256_GCM_OPENSSH_COM,
                'unparsable-algo',
            ],
            encryption_algorithms_server_to_client=[
                SshEncryptionAlgorithm.AES256_GCM_OPENSSH_COM,
                SshEncryptionAlgorithm.AES128_CBC,
                'unparsable-algo',
            ],
            mac_algorithms_client_to_server=[
                SshMacAlgorithm.HMAC_SHA1,
                SshMacAlgorithm.UMAC_128_OPENSSH_COM,
                'unparsable-algo',
            ],
            mac_algorithms_server_to_client=[
                SshMacAlgorithm.UMAC_128_OPENSSH_COM,
                SshMacAlgorithm.HMAC_SHA1,
                'unparsable-algo',
            ],
            compression_algorithms_client_to_server=[
                SshCompressionAlgorithm.NONE,
                SshCompressionAlgorithm.ZLIB_OPENSSH_COM,
                'unparsable-algo',
            ],
            compression_algorithms_server_to_client=[
                SshCompressionAlgorithm.ZLIB_OPENSSH_COM,
                SshCompressionAlgorithm.NONE,
                'unparsable-algo',
            ],
            languages_client_to_server=[
                LanguageTag('en', ['UK', ]),
                LanguageTag('en', ['US', ]),
            ],
            languages_server_to_client=[
                LanguageTag('en', ['US', ]),
                LanguageTag('en', ['UK', ]),
            ],
            cookie=b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f',
            reserved=0x00010203,
        )

    def test_parse(self):
        SshMessageVariantInit.parse_exact_size(self.key_exchange_init_bytes)

    def test_compose(self):
        self.assertEqual(self.key_exchange_init.compose(), self.key_exchange_init_bytes)

    def test_hassh(self):
        self.assertEqual(self.key_exchange_init.hassh, 'cc40dc455f685d8f57f8794262e30422')
        self.assertEqual(self.key_exchange_init.hassh_server, '55a954fe89f7f04218e3013996beee76')


class TestUnimplementedMessage(unittest.TestCase):
    def setUp(self):
        self.unimplemented_bytes = bytes(
            b'\x03' +                           # message_code = SshMessageCode.UNIMPLEMENTED
            b'\x01\x02\x03\x04' +               # sequence_number
            b''
        )
        self.unimplemented = SshUnimplementedMessage(
            sequence_number=0x01020304
        )

    def test_parse(self):
        message = SshMessageVariantInit.parse_exact_size(self.unimplemented_bytes)
        self.assertEqual(message.sequence_number, 0x01020304)

    def test_compose(self):
        self.assertEqual(self.unimplemented.compose(), self.unimplemented_bytes)


class TestDHKeyExchangeInit(unittest.TestCase):
    def setUp(self):
        self.dh_key_exchange_init_bytes = bytes(
            b'\x1e' +                              # message_code = SshMessageCode.DH_KEX_INIT
            b'\x00\x00\x00\x10' +                  # ephemeral_public_key length
            b'\x00\x01\x02\x03\x04\x05\x06\x07' +  # ephemeral_public_key length
            b'\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f' +
            b''
        )
        self.dh_key_exchange_init = SshDHKeyExchangeInit(
            ephemeral_public_key=b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f'
        )

    def test_parse(self):
        message = SshMessageVariantKexDH.parse_exact_size(self.dh_key_exchange_init_bytes)
        self.assertEqual(message.ephemeral_public_key, self.dh_key_exchange_init.ephemeral_public_key)

    def test_compose(self):
        self.assertEqual(self.dh_key_exchange_init.compose(), self.dh_key_exchange_init_bytes)


class TestDHGroupExchangeInit(unittest.TestCase):
    def setUp(self):
        self.dh_group_exchange_init_bytes = bytes(
            b'\x20' +                              # message_code = SshMessageCode.DH_KEX_INIT
            b'\x00\x00\x00\x10' +                  # ephemeral_public_key length
            b'\x00\x01\x02\x03\x04\x05\x06\x07' +  # ephemeral_public_key length
            b'\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f' +
            b''
        )
        self.dh_group_exchange_init = SshDHGroupExchangeInit(
            ephemeral_public_key=b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f'
        )

    def test_parse(self):
        message = SshMessageVariantKexDHGroup.parse_exact_size(self.dh_group_exchange_init_bytes)
        self.assertEqual(message, self.dh_group_exchange_init)

    def test_compose(self):
        self.assertEqual(self.dh_group_exchange_init.compose(), self.dh_group_exchange_init_bytes)


class TestDHKeyExchangeReply(unittest.TestCase):
    def setUp(self):
        self.dh_key_exchange_reply_dict = collections.OrderedDict([
            ('message_code', b'\x1f'),  # DH_KEX_REPLY
            ('host_key_length', b'\x00\x00\x00\x23'),
            ('host_public_key', (
                b'\x00\x00\x00\x07' +
                b'ssh-rsa' +
                b'\x00\x00\x00\x08' +
                b'\x01\x01\x02\x03\x04\x05\x06\x07' +
                b'\x00\x00\x00\x08' +
                b'\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f' +
                b''
            )),
            ('ephemeral_public_key_length', b'\x00\x00\x00\x10'),
            ('ephemeral_public_key', (
                b'\x00\x01\x02\x03\x04\x05\x06\x07' +
                b'\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f' +
                b''
            )),
            ('signature_length', b'\x00\x00\x00\x10'),
            ('signature_key', (
                b'\x00\x01\x02\x03\x04\x05\x06\x07' +
                b'\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f' +
                b''
            )),
        ])
        self.dh_key_exchange_reply_bytes = b''.join(self.dh_key_exchange_reply_dict.values())
        self.dh_key_exchange_reply = SshDHKeyExchangeReply(
            host_public_key=SshHostKeyRSA(
                host_key_algorithm=SshHostKeyAlgorithm.SSH_RSA,
                public_key=PublicKey.from_params(PublicKeyParamsRsa(
                    modulus=0x08090a0b0c0d0e0f,
                    public_exponent=0x0101020304050607,
                )),
            ),
            ephemeral_public_key=b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f',
            signature=b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f',
        )

    def test_parse(self):
        message = SshMessageVariantKexDH.parse_exact_size(self.dh_key_exchange_reply_bytes)
        self.assertEqual(message, self.dh_key_exchange_reply)

    def test_compose(self):
        self.assertEqual(self.dh_key_exchange_reply.compose(), self.dh_key_exchange_reply_bytes)


class TestDHGroupExchangeReply(unittest.TestCase):
    def setUp(self):
        self.dh_group_exchange_reply_dict = collections.OrderedDict([
            ('message_code', b'\x21'),  # DH_GEX_REPLY
            ('host_key_length', b'\x00\x00\x00\x23'),
            ('host_public_key', (
                b'\x00\x00\x00\x07' +
                b'ssh-rsa' +
                b'\x00\x00\x00\x08' +
                b'\x01\x01\x02\x03\x04\x05\x06\x07' +
                b'\x00\x00\x00\x08' +
                b'\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f' +
                b''
            )),
            ('ephemeral_public_key_length', b'\x00\x00\x00\x10'),
            ('ephemeral_public_keylic_key', (
                b'\x00\x01\x02\x03\x04\x05\x06\x07' +
                b'\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f' +
                b''
            )),
            ('signature_length', b'\x00\x00\x00\x10'),
            ('signature_key', (
                b'\x00\x01\x02\x03\x04\x05\x06\x07' +
                b'\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f' +
                b''
            )),
        ])
        self.dh_group_exchange_reply_bytes = b''.join(self.dh_group_exchange_reply_dict.values())
        self.dh_group_exchange_reply = SshDHGroupExchangeReply(
            host_public_key=SshHostKeyRSA(
                host_key_algorithm=SshHostKeyAlgorithm.SSH_RSA,
                public_key=PublicKey.from_params(PublicKeyParamsRsa(
                    modulus=0x08090a0b0c0d0e0f,
                    public_exponent=0x0101020304050607,
                )),
            ),
            ephemeral_public_key=b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f',
            signature=b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f',
        )

    def test_parse(self):
        message = SshMessageVariantKexDHGroup.parse_exact_size(self.dh_group_exchange_reply_bytes)
        self.assertEqual(message, self.dh_group_exchange_reply)

    def test_compose(self):
        self.assertEqual(self.dh_group_exchange_reply.compose(), self.dh_group_exchange_reply_bytes)


class TestDHGroupExchangeRequest(unittest.TestCase):
    def setUp(self):
        self.dh_group_exchange_reply_dict = collections.OrderedDict([
            ('message_code', b'\x22'),  # DH_GEX_REQUEST
            ('gex_min', b'\x00\x00\x04\x00'),
            ('gex_number', b'\x00\x00\x08\x00'),
            ('gex_max', b'\x00\x00\x10\x00'),
        ])
        self.dh_group_exchange_reply_bytes = b''.join(self.dh_group_exchange_reply_dict.values())
        self.dh_group_exchange_reply = SshDHGroupExchangeRequest(
            gex_min=1024,
            gex_number=2048,
            gex_max=4096,
        )

    def test_parse(self):
        message = SshMessageVariantKexDHGroup.parse_exact_size(self.dh_group_exchange_reply_bytes)
        self.assertEqual(message, self.dh_group_exchange_reply)

    def test_compose(self):
        self.assertEqual(self.dh_group_exchange_reply.compose(), self.dh_group_exchange_reply_bytes)


class TestDHGroupExchangeGroup(unittest.TestCase):
    def setUp(self):
        self.dh_group_exchange_group_dict = collections.OrderedDict([
            ('message_code', b'\x1f'),  # DH_GEX_GROUP
            ('p_length', b'\x00\x00\x00\x10'),
            ('p', (
                b'\x00\x01\x02\x03\x04\x05\x06\x07' +
                b'\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f' +
                b''
            )),
            ('g_length', b'\x00\x00\x00\x10'),
            ('g', (
                b'\x00\x01\x02\x03\x04\x05\x06\x07' +
                b'\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f' +
                b''
            )),
        ])
        self.dh_group_exchange_group_bytes = b''.join(self.dh_group_exchange_group_dict.values())
        self.dh_group_exchange_group = SshDHGroupExchangeGroup(
            p=b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f',
            g=b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f',
        )

    def test_parse(self):
        message = SshMessageVariantKexDHGroup.parse_exact_size(self.dh_group_exchange_group_bytes)
        self.assertEqual(message, self.dh_group_exchange_group)

    def test_compose(self):
        self.assertEqual(self.dh_group_exchange_group.compose(), self.dh_group_exchange_group_bytes)


class TestNewKeys(unittest.TestCase):
    def setUp(self):
        self.new_keys_dict = collections.OrderedDict([
            ('message_code', b'\x15'),  # NEWKEYS
        ])
        self.new_keys_bytes = b''.join(self.new_keys_dict.values())
        self.new_keys = SshNewKeys()

    def test_parse(self):
        message = SshNewKeys.parse_exact_size(self.new_keys_bytes)
        self.assertEqual(message, self.new_keys)

    def test_compose(self):
        self.assertEqual(self.new_keys.compose(), self.new_keys_bytes)



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import unittest

from cryptodatahub.common.exception import InvalidValue

from cryptoparser.common.exception import NotEnoughData

from cryptoparser.tls.mysql import (
    MySQLRecord,
    MySQLCapability,
    MySQLCharacterSet,
    MySQLStatusFlag,
    MySQLHandshakeSslRequest,
    MySQLHandshakeV10,
    MySQLVersion,
)


class TestMySQLRecord(unittest.TestCase):
    def setUp(self):
        self.test_record = MySQLRecord(
            packet_number=1,
            packet_bytes=b'\x01\x02\x03\x04'
        )
        self.test_record_bytes = bytes(
            b'\x04\x00\x00' +    # packet_length
            b'\x01' +            # packet_number
            b'\x01\x02\x03\x04'
        )

    def test_error(self):
        with self.assertRaises(NotEnoughData) as context_manager:
            MySQLRecord.parse_exact_size(b'\x00')
        self.assertEqual(context_manager.exception.bytes_needed, MySQLRecord.HEADER_SIZE - 1)

    def test_parse(self):
        MySQLRecord.parse_exact_size(self.test_record_bytes)

    def test_compose(self):
        self.assertEqual(self.test_record.compose(), self.test_record_bytes)


class TestMySQLHandshake10(unittest.TestCase):
    def setUp(self):
        self.handshake_minimal = MySQLHandshakeV10(
            protocol_version=MySQLVersion.MYSQL_9,
            server_version='1.2.3.4',
            connection_id=0x01020304,
            auth_plugin_data=b'\x01\x02\x03\x04\x05\x06\x07\x08',
            capabilities={
                MySQLCapability.CLIENT_SSL,
                MySQLCapability.CLIENT_MULTI_STATEMENTS,
            },
            character_set=MySQLCharacterSet.UTF8,
            states={MySQLStatusFlag.SERVER_STATUS_IN_TRANS, },
        )
        self.handshake_full = MySQLHandshakeV10(
            protocol_version=MySQLVersion.MYSQL_9,
            server_version='1.2.3.4',
            connection_id=0x01020304,
            auth_plugin_data=b'\x01\x02\x03\x04\x05\x06\x07\x08',
            capabilities={
                MySQLCapability.CLIENT_SSL,
                MySQLCapability.CLIENT_MULTI_STATEMENTS,
                MySQLCapability.CLIENT_PLUGIN_AUTH,
            },
            character_set=MySQLCharacterSet.UTF8,
            states={MySQLStatusFlag.SERVER_STATUS_IN_TRANS, },
            auth_plugin_name='auth_plugin_name',
            auth_plugin_data_2=b'\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d'
        )

        self.handshake_bytes_base = bytes(
            b'\x09' +                              # protocol_version
            b'1.2.3.4\x00' +                       # server_version
            b'\x04\x03\x02\x01' +                  # connection_id
            b'\x01\x02\x03\x04\x05\x06\x07\x08' +  # auth_plugin_data
            b'\x00' +                              # filler
            b'\x00\x08' +                          # capabilities
            b'\x21' +                              # character_set
            b'\x01\x00' +                          # states
            b''
        )
        self.handshake_bytes_minimal = self.handshake_bytes_base + bytes(
            b'\x01\x00' +                          # capabilities_2
            b'\x00' +                              # auth_plugin_data_len
            10 * b'\x00' +                         # reserved
            b''
        )
        self.handshake_bytes_full = self.handshake_bytes_base + bytes(
            b'\x09\x00' +                          # capabilities_2
            b'\x15' +                              # auth_plugin_data_len
            10 * b'\x00' +                         # reserved
            b'\x01\x02\x03\x04\x05\x06\x07\x08' +  # auth_plugin_data_2
            b'\x09\x0a\x0b\x0c\x0d' +
            b'auth_plugin_name\x00' +              # auth_plugin_name
            b''
        )

    def test_error_not_enough_data(self):
        with self.assertRaises(NotEnoughData) as context_manager:
            MySQLHandshakeV10.parse_exact_size(b'\x00')
        self.assertEqual(context_manager.exception.bytes_needed, MySQLHandshakeV10.MINIMUM_SIZE - 1)

    def test_error_no_auth_plugin_data(self):
        handshake_bytes_no_auth_plugin_data = self.handshake_bytes_base + bytes(
            b'\x09\x00' +                          # capabilities_2
            b'\x00' +                              # auth_plugin_data_len
            10 * b'\x00' +                         # reserved
            b''
        )
        with self.assertRaises(InvalidValue) as context_manager:
            MySQLHandshakeV10.parse_exact_size(handshake_bytes_no_auth_plugin_data)
        self.assertEqual(context_manager.exception.value, 0)

    def test_parse(self):
        handshake_minimal = MySQLHandshakeV10.parse_exact_size(self.handshake_bytes_minimal)
        self.assertEqual(handshake_minimal, self.handshake_minimal)

        handshake_full = MySQLHandshakeV10.parse_exact_size(self.handshake_bytes_full)
        self.assertEqual(handshake_full, self.handshake_full)

    def test_compose(self):
        self.assertEqual(self.handshake_minimal.compose(), self.handshake_bytes_minimal)
        self.assertEqual(self.handshake_full.compose(), self.handshake_bytes_full)


class TestMySQLHandshakeSslRequest(unittest.TestCase):
    def setUp(self):
        self.handshake_bytes_with_client_41_capability = bytes(
            b'\x00\x02\x00\x00' +                  # capabilities
            b'\x04\x03\x02\x01' +                  # max_packet_size
            b'\x21' +                              # character_set
            23 * b'\x00' +                         # filler
            b''
        )
        self.handshake_with_client_41_capability = MySQLHandshakeSslRequest(
            capabilities={MySQLCapability.CLIENT_PROTOCOL_41},
            max_packet_size=0x01020304,
            character_set=MySQLCharacterSet.UTF8
        )
        self.handshake_bytes_without_client_41_capability = bytes(
            b'\x00\x00' +                  # capabilities
            b'\x03\x02\x01' +              # max_packet_size
            b''
        )
        self.handshake_without_client_41_capability = MySQLHandshakeSslRequest(
            capabilities=set(),
            max_packet_size=0x010203,
        )

    def test_error_not_enough_data(self):
        with self.assertRaises(NotEnoughData) as context_manager:
            MySQLHandshakeSslRequest.parse_exact_size(b'\x00')
        self.assertEqual(context_manager.exception.bytes_needed, MySQLHandshakeSslRequest.MINIMUM_SIZE - 1)

    def test_with_client_41_character_set_is_none(self):
        ssl_request = MySQLHandshakeSslRequest(
            capabilities=set([MySQLCapability.CLIENT_PROTOCOL_41]), max_packet_size=1, character_set=None
        )
        self.assertEqual(ssl_request.character_set, MySQLCharacterSet.UTF8)

    def test_error_without_client_41_capability_too_large(self):
        with self.assertRaises(ValueError) as context_manager:
            MySQLHandshakeSslRequest(capabilities=set([MySQLCapability.CLIENT_MULTI_STATEMENTS]), max_packet_size=1)
        self.assertEqual(context_manager.exception.args[0], 1)

    def test_error_without_client_41_max_packet_size_too_large(self):  # pylint: disable=invalid-name
        with self.assertRaises(ValueError) as context_manager:
            MySQLHandshakeSslRequest(capabilities=set(), max_packet_size=2 ** 24)
        self.assertEqual(context_manager.exception.args[0], 2 ** 24)

    def test_with_client_41_capability(self):
        self.assertEqual(
            MySQLHandshakeSslRequest.parse_exact_size(self.handshake_bytes_with_client_41_capability),
            self.handshake_with_client_41_capability,
        )
        self.assertEqual(
            self.handshake_with_client_41_capability.compose(),
            self.handshake_bytes_with_client_41_capability,
        )

    def test_without_client_41_capability(self):
        self.assertEqual(
            MySQLHandshakeSslRequest.parse_exact_size(self.handshake_bytes_without_client_41_capability),
            self.handshake_without_client_41_capability,
        )
        self.assertEqual(
            self.handshake_without_client_41_capability.compose(),
            self.handshake_bytes_without_client_41_capability,
        )



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import abc

from cryptodatahub.tls.algorithm import TlsECPointFormat, TlsNamedCurve, TlsSignatureAndHashAlgorithm

from cryptoparser.common.base import OneByteEnumParsable, TwoByteEnumParsable


class TlsNamedCurveFactory(TwoByteEnumParsable):
    @classmethod
    def get_enum_class(cls):
        return TlsNamedCurve

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()


class TlsSignatureAndHashAlgorithmFactory(TwoByteEnumParsable):
    @classmethod
    def get_enum_class(cls):
        return TlsSignatureAndHashAlgorithm

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()


class TlsECPointFormatFactory(OneByteEnumParsable):
    @classmethod
    def get_enum_class(cls):
        return TlsECPointFormat

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import abc
import base64
import collections
import enum
import json
import pathlib

import attr

import pyfakefs.fake_filesystem_unittest

from cryptodatahub.common.exception import InvalidValue
from cryptodatahub.common.grade import AttackNamed, AttackType, Grade, GradeableVulnerabilities, Vulnerability
from cryptodatahub.common.types import CryptoDataEnumBase, CryptoDataParamsNamed

from cryptoparser.common.base import (
    ComposerBinary,
    FourByteEnumComposer,
    FourByteEnumParsable,
    ListParamParsable,
    ListParsable,
    NumericRangeParsableBase,
    OneByteEnumComposer,
    OneByteEnumParsable,
    OpaqueEnumComposer,
    OpaqueEnumParsable,
    OpaqueParam,
    ParsableBase,
    ParserBinary,
    Serializable,
    SerializableTextEncoder,
    StringEnumCaseInsensitiveParsable,
    StringEnumParsable,
    ThreeByteEnumComposer,
    ThreeByteEnumParsable,
    TwoByteEnumComposer,
    TwoByteEnumParsable,
    VariantParsable,
    VariantParsableExact,
)
from cryptoparser.common.exception import TooMuchData, InvalidType
from cryptoparser.common.field import (
    FieldsJson,
    FieldsSemicolonSeparated,
    FieldValueStringEnum,
    FieldValueComponentBool,
    FieldValueComponentFloat,
    FieldValueComponentNumber,
    FieldValueComponentOption,
    FieldValueComponentPercent,
    FieldValueComponentQuotedString,
    FieldValueComponentString,
    FieldValueComponentStringBase64,
    FieldValueComponentStringEnum,
    FieldValueComponentStringEnumParams,
    FieldValueComponentTimeDelta,
    FieldValueComponentUrl,
    FieldValueStringEnumParams,
    FieldValueTimeDelta,
    NameValuePairListSemicolonSeparated,
)
from cryptoparser.common.parse import ParserCRLF
from cryptoparser.common.x509 import PublicKeyX509


class NByteParsable(ParsableBase):
    def __init__(self, value):
        if value < 0 or value >= 2 ** (8 * self.get_byte_size()):
            raise ValueError

        self.value = value

    def __int__(self):
        return self.value

    @classmethod
    def _parse(cls, parsable):
        parser = ParserBinary(parsable)

        parser.parse_numeric('value', cls.get_byte_size())

        return cls(parser['value']), cls.get_byte_size()

    def compose(self):
        composer = ComposerBinary()

        composer.compose_numeric(self.value, self.get_byte_size())

        return composer.composed_bytes

    def __repr__(self):
        return f'0x{self.value:>0{self.get_byte_size() * 2}x}'

    def __eq__(self, other):
        return self.get_byte_size() == other.get_byte_size() and self.value == other.value

    @classmethod
    def get_byte_size(cls):
        raise NotImplementedError()


class OneByteParsable(NByteParsable):
    @classmethod
    def get_byte_size(cls):
        return 1


class TwoByteParsable(NByteParsable):
    @classmethod
    def get_byte_size(cls):
        return 2


class ConditionalParsable(NByteParsable):
    def __int__(self):
        return self.value

    @classmethod
    def _parse(cls, parsable):
        parser = ParserBinary(parsable)

        parser.parse_numeric('value', cls.get_byte_size())

        cls.check_parsed(parser['value'])

        return cls(parser['value']), cls.get_byte_size()

    def compose(self):
        composer = ComposerBinary()

        composer.compose_numeric(self.value, self.get_byte_size())

        return composer.composed_bytes

    @classmethod
    @abc.abstractmethod
    def check_parsed(cls, value):
        raise NotImplementedError()


class OneByteOddParsable(ConditionalParsable):
    @classmethod
    def get_byte_size(cls):
        return 1

    @classmethod
    def check_parsed(cls, value):
        if value % 2 == 0:
            raise InvalidValue(value, OneByteOddParsable)


class TwoByteEvenParsable(ConditionalParsable):
    @classmethod
    def get_byte_size(cls):
        return 2

    @classmethod
    def check_parsed(cls, value):
        if value % 2 != 0:
            raise InvalidValue(value, TwoByteEvenParsable)


class AlwaysUnknowTypeParsable(ParsableBase):
    @classmethod
    def _parse(cls, parsable):
        raise InvalidValue(parsable, AlwaysUnknowTypeParsable)

    def compose(self):
        raise TooMuchData()


class AlwaysInvalidTypeParsable(ParsableBase):
    @classmethod
    def _parse(cls, parsable):
        raise InvalidType()

    def compose(self):
        raise TooMuchData()


class AlwaysInvalidTypeVariantParsable(VariantParsable):
    @classmethod
    def _get_variants(cls):
        return collections.OrderedDict([
            (AlwaysInvalidTypeParsable, (AlwaysInvalidTypeParsable, ))
        ])


class SerializableEnumVariantParsable(VariantParsable):
    @classmethod
    def _get_variants(cls):
        return collections.OrderedDict([
            (SerializableEnum, (SerializableEnum, ))
        ])


class AlwaysTestStringComposer(ParsableBase):
    def __eq__(self, other):
        return isinstance(other, AlwaysTestStringComposer)

    @classmethod
    def _parse(cls, parsable):
        if parsable[:4] != b'test':
            raise InvalidValue(parsable, cls)

        return AlwaysTestStringComposer(), 4

    def compose(self):
        return b'test'


@attr.s
class SerializableEnumValue(Serializable):
    code = attr.ib(validator=attr.validators.instance_of(int))

    def as_json(self):
        return json.dumps({'code': self.code})

    def _as_markdown(self, level):
        return False, self.code

    @classmethod
    def get_code_size(cls):
        return 2


class OpaqueEnumFactory(OpaqueEnumParsable):
    @classmethod
    def get_enum_class(cls):
        return OpaqueEnum

    @classmethod
    def get_param(cls):
        return OpaqueParam(
            min_byte_num=1, max_byte_num=32
        )


@attr.s
class OpaqueEnumParams():
    code = attr.ib(validator=attr.validators.instance_of(str))


class OpaqueEnum(OpaqueEnumComposer):
    ALPHA = OpaqueEnumParams(code='άλφα')
    BETA = OpaqueEnumParams(code='βήτα')
    GAMMA = OpaqueEnumParams(code='γάμμα')


class TestObject():
    pass


class SerializableSimpleTypes(Serializable):  # pylint: disable=too-many-instance-attributes
    def __init__(self):
        self.int_value = 1
        self.float_value = 1.0
        self.bool_value = False
        self.str_value = 'string'
        self.bytearray_value = bytearray(b'\x00\x01\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f')
        self.none_value = None


class SerializableIterables(Serializable):
    def __init__(self):
        self.dict_key = collections.OrderedDict([
            (1, 'int'),
            ('str', 'string'),
            (SerializableStringEnum.FIRST, 'enum'),
        ])
        self.dict_value = collections.OrderedDict([
            ('int', 1),
            ('string', 'str'),
            ('enum', SerializableStringEnum.FIRST),
        ])
        self.list_value = list(['value', ])
        self.tuple_value = tuple(['value', ])


class SerializableEnumFactory(TwoByteEnumParsable):
    @classmethod
    def get_enum_class(cls):
        return SerializableEnum

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()


class SerializableEnum(TwoByteEnumComposer):
    FIRST = SerializableEnumValue(
        code=0x0001,
    )
    SECOND = SerializableEnumValue(
        code=0x0002,
    )


class SerializableStringEnum(enum.Enum):
    FIRST = '1'
    SECOND = '2'


class SerializableEnums(Serializable):
    def __init__(self):
        self.param_enum = SerializableEnum.FIRST
        self.string_enum = SerializableStringEnum.SECOND


class SerializableHidden(Serializable):
    def __init__(self):
        self._invisible_value = None
        self.visible_value = 'value'


class SerializableSingle(Serializable):
    def _asdict(self):
        return 'single'


class SerializableUnhandled(Serializable):
    def __init__(self):
        self.complex_number = 1 + 2j


@attr.s
class SerializableHumanReadable(Serializable):
    attr_2 = attr.ib(default='value 2', metadata={'human_readable_name': 'Human Readable Name 2'})
    attr_1 = attr.ib(default='value 1', metadata={'human_readable_name': 'Human Readable Name 1'})


@attr.s
class SerializableHumanFriendly(Serializable):
    human_friendly = attr.ib(default='human-friendly', metadata={'human_friendly': True})
    human_friendly_by_default = attr.ib(default='human-friendly-by-default')
    non_human_friendly = attr.ib(default='non-human-friendly', metadata={'human_friendly': False})


@attr.s
class SerializableAttributeOrder(Serializable):
    attr_b = attr.ib(default='b')
    attr_a = attr.ib(default='a')


class Class():
    def __init__(self):
        self.attr_b = 'b'
        self.attr_a = 'a'


@attr.s
class ClassAttr():
    attr_b = attr.ib(default='b')
    attr_a = attr.ib(default='a')


class ClassAsDict():
    def __init__(self):
        self.attr_a = 'a'
        self.attr_b = 'b'

    def _asdict(self):
        return collections.OrderedDict([
            ('attr_b', self.attr_b),
            ('attr_a', self.attr_a),
        ])


@attr.s
class ClassAttrAsDict():
    attr_a = attr.ib(default='a')
    attr_b = attr.ib(default='b')

    def _asdict(self):
        return collections.OrderedDict([
            ('attr_b', self.attr_b),
            ('attr_a', self.attr_a),
        ])


class ClassCryptoDataEnum(CryptoDataEnumBase):
    ONE = CryptoDataParamsNamed('one', 'long one')


class ClassGradeable(GradeableVulnerabilities):
    @classmethod
    def get_gradeable_name(cls):
        return 'gradeable'

    def __str__(self):
        return 'value'


class SerializableRecursive(Serializable):  # pylint: disable=too-many-instance-attributes
    def __init__(self):
        self.json_object = Class()
        self.json_attr_object = ClassAttr()
        self.json_attr_as_dict = ClassAttrAsDict()
        self.json_asdict_object = ClassAsDict()
        self.json_crypto_data_hub_enum = ClassCryptoDataEnum.ONE
        self.json_gradeable = ClassGradeable([
            Vulnerability(AttackType.MITM, Grade.INSECURE, AttackNamed.NOFS)
        ])
        self.json_serializable_hidden = SerializableHidden()
        self.json_serializable_single = 'single'
        self.json_serializable_in_list = list([SerializableHidden(), 'single'])
        self.json_serializable_in_tuple = tuple([SerializableHidden(), 'single'])
        self.json_serializable_in_dict = dict({'key1': SerializableHidden(), 'key2': 'single'})


class SerializableEmptyValues(Serializable):
    def __init__(self):
        self.value = None
        self.list = []
        self.tuple = tuple()
        self.dict = {}


class FlagEnum(enum.IntEnum):
    ONE = 1
    TWO = 2
    FOUR = 4
    EIGHT = 8


@attr.s
class StringEnumParams():
    code = attr.ib()

    def _check_code(self, code):
        if code != self.code:
            raise InvalidType()

    @classmethod
    def get_code_size(cls):
        return 2


class StringEnum(StringEnumParsable, enum.Enum):
    ONE = StringEnumParams(
        code='one',
    )
    TWO = StringEnumParams(
        code='two',
    )
    THREE = StringEnumParams(
        code='three',
    )


class StringEnumA(StringEnumParsable, enum.Enum):
    A = StringEnumParams(code='a')


class StringEnumAA(StringEnumParsable, enum.Enum):
    AA = StringEnumParams(code='aa')


class StringEnumAAA(StringEnumParsable, enum.Enum):
    AAA = StringEnumParams(code='aaa')


class VariantParsableTest(VariantParsable):
    @classmethod
    def _get_variants(cls):
        return collections.OrderedDict([
            (StringEnumA, [StringEnumA, ]),
            (StringEnumAA, [StringEnumAA, ]),
            (StringEnumAAA, [StringEnumAAA, ]),
        ])


class VariantParsableExactTest(VariantParsableExact):
    @classmethod
    def _get_variants(cls):
        return collections.OrderedDict([
            (StringEnumA, [StringEnumA, ]),
            (StringEnumAA, [StringEnumAA, ]),
            (StringEnumAAA, [StringEnumAAA, ]),
        ])


class EnumStringValue(enum.Enum):
    ONE = 'one'
    TWO = 'two'
    THREE = 'three'


class ListParamParsableTest(ListParamParsable):
    pass


class ListParsableTest(ListParsable):
    @classmethod
    def get_param(cls):
        return ListParamParsableTest(
            item_class=AlwaysTestStringComposer,
            fallback_class=None,
            separator_class=ParserCRLF,
        )


@attr.s
class NByteEnumParam():
    code = attr.ib(validator=attr.validators.instance_of(int))


class NByteEnumTest(enum.Enum):
    ONE = NByteEnumParam(code=1)
    TWO = NByteEnumParam(code=2)
    THREE = NByteEnumParam(code=3)
    FOUR = NByteEnumParam(code=4)


class OneByteEnumParsableTest(OneByteEnumParsable):
    @classmethod
    def get_enum_class(cls):
        return NByteEnumTest

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()


class OneByteEnumComposerTest(OneByteEnumComposer, enum.Enum):
    ONE = NByteEnumParam(code=1)
    TWO = NByteEnumParam(code=2)
    THREE = NByteEnumParam(code=3)
    FOUR = NByteEnumParam(code=4)


class TwoByteEnumParsableTest(TwoByteEnumParsable):
    @classmethod
    def get_enum_class(cls):
        return NByteEnumTest

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()


class TwoByteEnumComposerTest(TwoByteEnumComposer, enum.Enum):
    ONE = NByteEnumParam(code=1)
    TWO = NByteEnumParam(code=2)
    THREE = NByteEnumParam(code=3)
    FOUR = NByteEnumParam(code=4)


class ThreeByteEnumParsableTest(ThreeByteEnumParsable):
    @classmethod
    def get_enum_class(cls):
        return NByteEnumTest

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()


class ThreeByteEnumComposerTest(ThreeByteEnumComposer, enum.Enum):
    ONE = NByteEnumParam(code=1)
    TWO = NByteEnumParam(code=2)
    THREE = NByteEnumParam(code=3)
    FOUR = NByteEnumParam(code=4)


class FourByteEnumParsableTest(FourByteEnumParsable):
    @classmethod
    def get_enum_class(cls):
        return NByteEnumTest

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()


class FourByteEnumComposerTest(FourByteEnumComposer, enum.Enum):
    ONE = NByteEnumParam(code=1)
    TWO = NByteEnumParam(code=2)
    THREE = NByteEnumParam(code=3)
    FOUR = NByteEnumParam(code=4)


class FieldValueTimeDeltaTest(FieldValueTimeDelta):
    @classmethod
    def get_name(cls):
        return 'testTimeDelta'


class FieldValueEnumTest(StringEnumCaseInsensitiveParsable, enum.Enum):
    FIRST = FieldValueStringEnumParams(code='first', human_readable_name='FiRsT')
    SECOND = FieldValueStringEnumParams(code='second')


class FieldValueStringEnumTest(FieldValueStringEnum):
    @classmethod
    def _get_value_type(cls):
        return FieldValueEnumTest


class FieldValueComponentOptionTest(FieldValueComponentOption):
    @classmethod
    def get_canonical_name(cls):
        return 'testOption'


class FieldValueComponentStringTest(FieldValueComponentString):
    @classmethod
    def get_canonical_name(cls):
        return 'testString'


class FieldValueComponentUrlTest(FieldValueComponentUrl):
    @classmethod
    def get_canonical_name(cls):
        return 'testUrl'


class ComponentStringEnumTest(StringEnumParsable, enum.Enum):
    ONE = FieldValueComponentStringEnumParams(
        code='one'
    )
    TWO = FieldValueComponentStringEnumParams(
        code='two'
    )
    THREE = FieldValueComponentStringEnumParams(
        code='three'
    )


class FieldValueComponentStringBase64Test(FieldValueComponentStringBase64):
    @classmethod
    def get_canonical_name(cls):
        return 'testStringBase64'


class FieldValueComponentBoolTest(FieldValueComponentBool):
    @classmethod
    def get_canonical_name(cls):
        return 'testBool'


class FieldValueComponentFloatTest(FieldValueComponentFloat):
    @classmethod
    def get_canonical_name(cls):
        return 'testFloat'


class FieldValueComponentStringEnumTest(FieldValueComponentStringEnum):
    @classmethod
    def get_canonical_name(cls):
        return 'testStringEnum'

    @classmethod
    def _get_value_type(cls):
        return ComponentStringEnumTest


class FieldValueComponentOptionalStringTest(FieldValueComponentQuotedString):
    @classmethod
    def get_canonical_name(cls):
        return 'testOptionalString'


class FieldValueComponentQuotedStringTest(FieldValueComponentQuotedString):
    @classmethod
    def get_canonical_name(cls):
        return 'testQuotedString'


class FieldValueComponentNumberTest(FieldValueComponentNumber):
    @classmethod
    def get_canonical_name(cls):
        return 'testNumber'


class FieldValueComponentPercentTest(FieldValueComponentPercent):
    @classmethod
    def get_canonical_name(cls):
        return 'testPercent'


class FieldValueComponentTimeDeltaTest(FieldValueComponentTimeDelta):
    @classmethod
    def get_canonical_name(cls):
        return 'testTimeDelta'


@attr.s
class FieldValueComplexTestBase():  # pylint: disable=too-many-instance-attributes
    time_delta = attr.ib(
        converter=FieldValueComponentTimeDeltaTest.convert,
        validator=attr.validators.instance_of(FieldValueComponentTimeDeltaTest)
    )
    string = attr.ib(
        converter=FieldValueComponentStringTest.convert,
        validator=attr.validators.instance_of(FieldValueComponentStringTest),
        default='default'
    )
    url = attr.ib(
        converter=FieldValueComponentUrlTest.convert,
        validator=attr.validators.instance_of(FieldValueComponentUrlTest),
        default='https://example.com'
    )
    base64_string = attr.ib(
        converter=FieldValueComponentStringBase64Test.convert,
        validator=attr.validators.instance_of(FieldValueComponentStringBase64Test),
        default=base64.b64encode('default'.encode('ascii')).decode('ascii')
    )
    number = attr.ib(
        converter=FieldValueComponentNumberTest.convert,
        validator=attr.validators.instance_of(FieldValueComponentNumberTest),
        default=0
    )
    percent = attr.ib(
        converter=FieldValueComponentPercentTest.convert,
        validator=attr.validators.instance_of(FieldValueComponentPercentTest),
        default=100
    )


@attr.s
class FieldValueMultipleTest(  # pylint: disable=too-many-instance-attributes
        FieldsSemicolonSeparated, FieldValueComplexTestBase):
    option = attr.ib(
        converter=FieldValueComponentOptionTest.convert,
        validator=attr.validators.instance_of(FieldValueComponentOptionTest),
        default=False
    )
    optional_string = attr.ib(
        converter=attr.converters.optional(FieldValueComponentOptionalStringTest.convert),
        validator=attr.validators.optional(
            attr.validators.instance_of(FieldValueComponentOptionalStringTest)
        ),
        default=None
    )


class FieldValueJsonTest(  # pylint: disable=too-many-instance-attributes
        FieldsJson, FieldValueComplexTestBase):
    pass


@attr.s
class FieldValueMultipleExtendableTest(FieldValueMultipleTest):
    extensions = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(NameValuePairListSemicolonSeparated)),
        metadata={'extension': True},
    )


class SerializableUpperCaseEncoder(SerializableTextEncoder):
    def __call__(self, obj, level):
        _, string_result = super().__call__(obj, level)

        return False, string_result.upper()


class TestClasses:
    class TestKeyBase(pyfakefs.fake_filesystem_unittest.TestCase):
        def setUp(self):
            self.setUpPyfakefs()

            self.__certs_dir = pathlib.PurePath(__file__).parent.parent / 'common' / 'certs'
            self.fs.add_real_directory(str(self.__certs_dir))

        def _get_pem_str(self, public_key_file_name):
            public_key_path = self.__certs_dir / public_key_file_name
            with open(str(public_key_path), 'r', encoding='ascii') as pem_file:
                return pem_file.read()

        def _get_public_key_x509(self, public_key_file_name):
            return PublicKeyX509.from_pem(self._get_pem_str(public_key_file_name))


class NumericRangeParsableTest(NumericRangeParsableBase):
    @classmethod
    def _get_value_min(cls):
        return 0x01

    @classmethod
    def _get_value_max(cls):
        return 0xfe

    @classmethod
    def _get_value_length(cls):
        return 1



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import binascii
import inspect


def get_leaf_classes(base_class):

    def _get_leaf_classes(base_class):
        subclasses = []

        if base_class.__subclasses__():
            for subclass in base_class.__subclasses__():
                subclasses += _get_leaf_classes(subclass)
        else:
            if not inspect.isabstract(base_class):
                return [base_class, ]

        return subclasses

    return _get_leaf_classes(base_class)


def bytes_to_hex_string(byte_array, separator='', lowercase=False):
    if lowercase:
        format_str = '{:02x}'
    else:
        format_str = '{:02X}'

    return separator.join([format_str.format(x) for x in bytes(byte_array)])


def bytes_from_hex_string(hex_string, separator=''):
    if separator:
        hex_string = ''.join(hex_string.split(separator))

    try:
        binary_data = binascii.a2b_hex(hex_string)
    except (TypeError, ValueError) as e:
        raise ValueError(*e.args) from e

    return binary_data



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

from cryptoparser.tls.extension import TlsExtensionUnusedData, TlsExtensionType
from cryptoparser.tls.subprotocol import TlsSubprotocolMessageBase, TlsHandshakeMessage, TlsHandshakeType


class TestMessage(TlsSubprotocolMessageBase):
    @classmethod
    def get_handshake_type(cls):
        raise NotImplementedError


class TestVariantMessage(TlsHandshakeMessage):
    @classmethod
    def get_handshake_type(cls):
        return TlsHandshakeType.SERVER_HELLO_DONE

    @classmethod
    def _parse(cls, parsable):
        raise NotImplementedError

    def compose(self):
        raise NotImplementedError


class TestUnusedDataExtension(TlsExtensionUnusedData):
    @classmethod
    def get_extension_type(cls):
        return TlsExtensionType.RENEGOTIATION_INFO



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import attr

from cryptodatahub.common.exception import InvalidValue

from cryptoparser.common.parse import ParsableBase, ParserBinary, ComposerBinary
from cryptoparser.common.exception import NotEnoughData


@attr.s
class Sync(ParsableBase):
    MESSAGE_SIZE = 1
    COMMAND = b'S'

    @classmethod
    def _parse(cls, parsable):
        parser = ParserBinary(parsable)

        parser.parse_raw('command', cls.MESSAGE_SIZE)
        if parser['command'] != cls.COMMAND:
            raise InvalidValue(parser['command'], cls, 'command')

        return cls(), cls.MESSAGE_SIZE

    def compose(self):
        composer = ComposerBinary()

        composer.compose_raw(self.COMMAND)

        return composer.composed_bytes


class SslRequest(ParsableBase):
    MESSAGE_SIZE = 8
    REQUEST_CODE = 80877103

    @classmethod
    def _parse(cls, parsable):
        if len(parsable) < cls.MESSAGE_SIZE:
            raise NotEnoughData(cls.MESSAGE_SIZE - len(parsable))

        parser = ParserBinary(parsable)

        parser.parse_numeric('length', 4)
        if parser['length'] != cls.MESSAGE_SIZE:
            raise InvalidValue(parser['length'], cls, 'length')

        parser.parse_numeric('request_code', 4)
        if parser['request_code'] != cls.REQUEST_CODE:
            raise InvalidValue(parser['request_code'], cls, 'request_code')

        return cls(), cls.MESSAGE_SIZE

    def compose(self):
        composer = ComposerBinary()

        composer.compose_numeric(self.MESSAGE_SIZE, 4)
        composer.compose_numeric(self.REQUEST_CODE, 4)

        return composer.composed_bytes



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import unittest

from cryptodatahub.ssh.algorithm import (
    SshCompressionAlgorithm,
    SshEncryptionAlgorithm,
    SshHostKeyAlgorithm,
    SshKexAlgorithm,
    SshMacAlgorithm,
)


class TestSshAlgorithm(unittest.TestCase):
    def test_str(self):
        self.assertEqual(str(SshEncryptionAlgorithm.ACSS_OPENSSH_ORG.value), 'acss@openssh.org')
        self.assertEqual(str(SshMacAlgorithm.CRYPTICORE_MAC_SSH_COM.value), 'crypticore-mac@ssh.com')
        self.assertEqual(str(SshKexAlgorithm.DIFFIE_HELLMAN_GROUP1_SHA1.value), 'diffie-hellman-group1-sha1')
        self.assertEqual(str(SshHostKeyAlgorithm.SSH_ED25519.value), 'ssh-ed25519')
        self.assertEqual(str(SshCompressionAlgorithm.ZLIB_OPENSSH_COM.value), 'zlib@openssh.com')


class TestSshAlgorithmMac(unittest.TestCase):
    def test_size(self):
        self.assertEqual(SshMacAlgorithm.HMAC_SHA2_256.value.size, 256)
        self.assertEqual(SshMacAlgorithm.HMAC_SHA2_256_96.value.size, 96)



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-
# pylint: disable=too-many-lines

import abc
import base64
import binascii
import collections
import datetime
import enum
import itertools
import textwrap

from collections import OrderedDict

import ipaddress
import attr


from cryptodatahub.common.algorithm import Authentication, Hash, NamedGroup
from cryptodatahub.common.exception import InvalidValue
from cryptodatahub.common.key import (
    PublicKey,
    PublicKeyParamsDsa,
    PublicKeyParamsEcdsa,
    PublicKeyParamsEddsa,
    PublicKeyParamsRsa,
    PublicKeySize,
)
from cryptodatahub.common.utils import hash_bytes
from cryptodatahub.ssh.algorithm import SshHostKeyAlgorithm, SshHostKeyType, SshEllipticCurveIdentifier

from cryptoparser.common.base import (
    FourByteEnumComposer,
    FourByteEnumParsable,
    Serializable,
    StringEnumParsable,
    VariantParsable,
    VectorParamParsable,
    VectorParamString,
    VectorParsable,
    VectorParsableDerived,
    VectorString,
)
from cryptoparser.common.exception import InvalidType, NotEnoughData
from cryptoparser.common.parse import ParsableBase, ParserBinary, ComposerBinary, ComposerText
from cryptoparser.common.x509 import PublicKeyX509


@attr.s
class SshPublicKeyBase():
    host_key_algorithm = attr.ib(
        converter=SshHostKeyAlgorithm,
        validator=attr.validators.instance_of(SshHostKeyAlgorithm)
    )
    public_key = attr.ib(
        validator=attr.validators.instance_of(PublicKey)
    )

    _HEADER_SIZE = 4

    @classmethod
    def get_host_key_algorithms(cls):
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def key_bytes(self):
        raise NotImplementedError()

    @classmethod
    def _fingerprint(cls, hash_type, key_bytes, prefix):
        digest = hash_bytes(hash_type, key_bytes)

        if hash_type == Hash.MD5:
            fingerprint = ':'.join(textwrap.wrap(binascii.hexlify(digest).decode('ascii'), 2))
        else:
            fingerprint = base64.b64encode(digest).decode('ascii')

        return ':'.join((prefix, fingerprint))

    @property
    def key_size(self):
        return PublicKeySize(self.public_key.key_type, self.public_key.key_size)

    @property
    def fingerprints(self):
        key_bytes = self.key_bytes
        return OrderedDict([
            (hash_type, self._fingerprint(hash_type, key_bytes, prefix))
            for hash_type, prefix in [(Hash.SHA2_256, 'SHA256'), (Hash.SHA1, 'SHA1'), (Hash.MD5, 'MD5')]
        ])

    def host_key_asdict(self):
        known_hosts = base64.b64encode(self.key_bytes).decode('ascii')

        public_key_dict = (
            [('key_type', self.host_key_algorithm.value.key_type.value)] +
            list(self.public_key._asdict().items()) +
            [('known_hosts', known_hosts)]
        )

        public_key_dict = OrderedDict(public_key_dict)
        public_key_dict['fingerprints'] = self.fingerprints

        return public_key_dict

    def _asdict(self):
        return self.host_key_asdict()

    @classmethod
    def _parse_host_key_algorithm(cls, parsable):
        if len(parsable) < cls._HEADER_SIZE:
            raise NotEnoughData(cls._HEADER_SIZE - len(parsable))

        parser = ParserBinary(parsable)
        parser.parse_string('host_key_algorithm', 4, 'ascii', SshHostKeyAlgorithm.from_code)

        if parser['host_key_algorithm'] not in cls.get_host_key_algorithms():
            raise InvalidType()

        return parser

    def _compose_host_key_algorithm(self):
        composer = ComposerBinary()

        host_key_algorithm_bytes = self.host_key_algorithm.value.code.encode('ascii')
        composer.compose_bytes(host_key_algorithm_bytes, 4)

        return composer


class SshHostKeyBase(SshPublicKeyBase):
    @classmethod
    @abc.abstractmethod
    def get_host_key_algorithms(cls):
        raise NotImplementedError()


class SshHostKeyParserBase(ParsableBase):
    @classmethod
    @abc.abstractmethod
    def _parse_host_key_algorithm(cls, parsable):
        raise NotImplementedError()

    @abc.abstractmethod
    def _compose_host_key_algorithm(self):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def _parse_host_key(cls, parser):
        raise NotImplementedError()

    @abc.abstractmethod
    def _compose_host_key_params(self, composer):
        raise NotImplementedError()

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_host_key_algorithm(parsable)
        host_key = cls._parse_host_key(parser)

        return cls(parser['host_key_algorithm'], host_key), parser.parsed_length

    def compose(self):
        composer = self._compose_host_key_algorithm()

        self._compose_host_key_params(composer)

        return composer.composed


@attr.s
class SshHostKeyDSSBase(SshHostKeyBase):
    @property
    @abc.abstractmethod
    def key_bytes(self):
        raise NotImplementedError()

    @classmethod
    def get_host_key_algorithms(cls):
        return filter(
            lambda host_key_algorithm: (
                host_key_algorithm.value.key_type == SshHostKeyType.HOST_KEY and
                host_key_algorithm.value.signature.value.key_type == Authentication.DSS
            ),
            SshHostKeyAlgorithm
        )

    @classmethod
    def _parse_host_key(cls, parser):
        for param_name in ['p', 'q', 'g', 'y']:
            parser.parse_ssh_mpint(param_name)

        public_key = PublicKey.from_params(PublicKeyParamsDsa(
            prime=parser['p'],
            generator=parser['g'],
            order=parser['q'],
            public_key_value=parser['y'],
        ))

        for param_name in ['p', 'q', 'g', 'y']:
            del parser[param_name]

        return public_key

    def _compose_host_key_params(self, composer):
        params = self.public_key.params
        for param_name in ['prime', 'order', 'generator', 'public_key_value']:
            value = getattr(params, param_name)
            composer.compose_ssh_mpint(value)

    def host_key_asdict(self):
        key_dict = OrderedDict([])

        key_dict.update(SshHostKeyBase.host_key_asdict(self))
        key_dict.update(OrderedDict([
            (param_name, getattr(self, param_name))
            for param_name in attr.fields_dict(SshHostKeyDSSBase).keys()
        ]))

        return key_dict


@attr.s
class SshHostKeyDSS(SshHostKeyDSSBase, SshHostKeyParserBase):
    @property
    def key_bytes(self):
        return self.compose()


@attr.s
class SshHostKeyRSABase(SshHostKeyBase):
    @property
    @abc.abstractmethod
    def key_bytes(self):
        raise NotImplementedError()

    @classmethod
    def get_host_key_algorithms(cls):
        return filter(
            lambda host_key_algorithm: (
                host_key_algorithm.value.key_type == SshHostKeyType.HOST_KEY and
                host_key_algorithm.value.signature.value.key_type == Authentication.RSA
            ),
            SshHostKeyAlgorithm
        )

    @classmethod
    def _parse_host_key(cls, parser):
        parser.parse_ssh_mpint('e')
        parser.parse_ssh_mpint('n')

        public_key = PublicKey.from_params(PublicKeyParamsRsa(
            modulus=parser['n'],
            public_exponent=parser['e'],
        ))

        del parser['e']
        del parser['n']

        return public_key

    def _compose_host_key_params(self, composer):
        params = self.public_key.params

        composer.compose_ssh_mpint(params.public_exponent)
        composer.compose_ssh_mpint(params.modulus)


@attr.s
class SshHostKeyRSA(SshHostKeyRSABase, SshHostKeyParserBase):
    @property
    def key_bytes(self):
        return self.compose()


@attr.s
class SshHostKeyECDSABase(SshHostKeyBase):
    @property
    @abc.abstractmethod
    def key_bytes(self):
        raise NotImplementedError()

    @classmethod
    def get_host_key_algorithms(cls):
        return filter(
            lambda host_key_algorithm: (
                host_key_algorithm.value.key_type == SshHostKeyType.HOST_KEY and
                host_key_algorithm.value.signature.value.key_type == Authentication.ECDSA
            ),
            SshHostKeyAlgorithm
        )

    @classmethod
    def _parse_host_key(cls, parser):
        parser.parse_string('curve_identifier', 4, 'ascii', SshEllipticCurveIdentifier.from_code)
        parser.parse_bytes('curve_data', 4)

        public_key = PublicKey.from_params(PublicKeyParamsEcdsa.from_octet_bit_string(
            parser['curve_identifier'].value.named_group,
            parser['curve_data'],
        ))

        del parser['curve_identifier']
        del parser['curve_data']

        return public_key

    def _compose_host_key_params(self, composer):
        named_group = self.public_key.params.named_group
        for elliptic_curve_identifier in SshEllipticCurveIdentifier:
            if elliptic_curve_identifier.value.named_group == named_group:
                composer.compose_string(elliptic_curve_identifier.value.code, 'ascii', 4)
                break
        else:
            raise NotImplementedError(named_group)

        composer.compose_bytes(self.public_key.params.octet_bit_string, 4)


@attr.s
class SshHostKeyECDSA(SshHostKeyECDSABase, SshHostKeyParserBase):
    @property
    def key_bytes(self):
        return self.compose()


@attr.s
class SshHostKeyEDDSABase(SshHostKeyBase):
    @property
    @abc.abstractmethod
    def key_bytes(self):
        raise NotImplementedError()

    @classmethod
    def get_host_key_algorithms(cls):
        return filter(
            lambda host_key_algorithm: (
                host_key_algorithm.value.key_type == SshHostKeyType.HOST_KEY and
                host_key_algorithm.value.signature.value.key_type == Authentication.EDDSA
            ),
            SshHostKeyAlgorithm
        )

    @classmethod
    def _parse_host_key(cls, parser):
        parser.parse_bytes('key_data', 4)

        public_key = PublicKey.from_params(PublicKeyParamsEddsa(
            curve_type=NamedGroup.CURVE25519,
            key_data=parser['key_data'],
        ))

        del parser['key_data']

        return public_key

    def _compose_host_key_params(self, composer):
        composer.compose_bytes(self.public_key.params.key_data, 4)


@attr.s
class SshHostKeyEDDSA(SshHostKeyEDDSABase, SshHostKeyParserBase):
    @property
    def key_bytes(self):
        return self.compose()


@attr.s(frozen=True)
class SshCertTypeParams(Serializable):
    code = attr.ib(validator=attr.validators.instance_of(int))
    name = attr.ib(validator=attr.validators.instance_of(str))

    def _as_markdown(self, level):
        return self._markdown_result(self.name, level)


class SshCertType(FourByteEnumComposer, enum.Enum):
    SSH_CERT_TYPE_USER = SshCertTypeParams(code=1, name='User')
    SSH_CERT_TYPE_HOST = SshCertTypeParams(code=2, name='Host')


class SshCertTypeFactory(FourByteEnumParsable):
    @classmethod
    def get_enum_class(cls):
        return SshCertType

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()


@attr.s
class SshCertSignature(ParsableBase):
    signature_type = attr.ib(validator=attr.validators.instance_of(SshHostKeyAlgorithm))
    signature_data = attr.ib(attr.validators.instance_of((bytes, bytearray)))

    @classmethod
    def _parse(cls, parsable):
        parser = ParserBinary(parsable)

        parser.parse_string('signature_type', 4, 'ascii', SshHostKeyAlgorithm.from_code)
        parser.parse_bytes('signature_data', 4)

        return cls(**parser), parser.parsed_length

    def compose(self):
        composer = ComposerBinary()

        composer.compose_string(self.signature_type.value.code, 'ascii', 4)
        composer.compose_bytes(self.signature_data, 4)

        return composer.composed


@attr.s
class SshString(ParsableBase):
    value = attr.ib(validator=attr.validators.instance_of(str))

    @classmethod
    def _parse(cls, parsable):
        parser = ParserBinary(parsable)

        parser.parse_string('value', 4, 'ascii')

        return cls(**parser), parser.parsed_length

    def compose(self):
        composer = ComposerBinary()

        composer.compose_string(self.value, 'ascii', 4)

        return composer.composed


class SshCertValidPrincipals(VectorParsable):
    @classmethod
    def get_param(cls):
        return VectorParamParsable(
            item_class=SshString,
            fallback_class=None,
            min_byte_num=0, max_byte_num=2 ** 32 - 4
        )


@attr.s(frozen=True)
class SshCertExtensionParam():
    code = attr.ib(validator=attr.validators.instance_of(str))
    critical = attr.ib(validator=attr.validators.instance_of(bool))


class SshCertExtensionName(StringEnumParsable, enum.Enum):
    FORCE_COMMAND = SshCertExtensionParam(
        code='force-command',
        critical=True,
    )
    SOURCE_ADDRESS = SshCertExtensionParam(
        code='source-address',
        critical=True,
    )
    NO_PRESENCE_REQUIRED = SshCertExtensionParam(
        code='no-presence-required',
        critical=False,
    )
    PERMIT_X11_FORWARDING = SshCertExtensionParam(
        code='permit-X11-forwarding',
        critical=False,
    )
    PERMIT_AGENT_FORWARDING = SshCertExtensionParam(
        code='permit-agent-forwarding',
        critical=False,
    )
    PERMIT_PORT_FORWARDING = SshCertExtensionParam(
        code='permit-port-forwarding',
        critical=False,
    )
    PERMIT_PTY = SshCertExtensionParam(
        code='permit-pty',
        critical=False,
    )
    PERMIT_USER_RC = SshCertExtensionParam(
        code='permit-user-rc',
        critical=False,
    )


class SshCertConstraintVector(VectorParsableDerived):
    @classmethod
    def get_param(cls):
        return VectorParamParsable(
            item_class=SshCertExtensionParsed,
            fallback_class=SshCertExtensionUnparsed,
            min_byte_num=0, max_byte_num=2 ** 32 - 1
        )


@attr.s
class SshCertExtensionBase(ParsableBase):
    @classmethod
    @abc.abstractmethod
    def _parse(cls, parsable):
        raise NotImplementedError()

    @abc.abstractmethod
    def compose(self):
        raise NotImplementedError()


@attr.s
class SshCertExtensionUnparsed(SshCertExtensionBase):
    extension_name = attr.ib(validator=attr.validators.instance_of(str))
    extension_data = attr.ib(validator=attr.validators.instance_of((bytes, bytearray)))

    @classmethod
    def _parse(cls, parsable):
        parser = ParserBinary(parsable)

        parser.parse_string('extension_name', 4, 'ascii')
        parser.parse_bytes('extension_data', 4)

        return SshCertExtensionUnparsed(**parser), parser.parsed_length

    def compose(self):
        payload_composer = ComposerBinary()

        payload_composer.compose_string(self.extension_name, 'ascii', 4)
        payload_composer.compose_bytes(self.extension_data, 4)

        return payload_composer.composed_bytes


@attr.s
class SshCertExtensionParsed(SshCertExtensionBase):
    extension_name = attr.ib(init=False, validator=attr.validators.instance_of(SshCertExtensionName))

    def __attrs_post_init__(self):
        self.extension_name = self.get_extension_name()

        attr.validate(self)

    @classmethod
    @abc.abstractmethod
    def get_extension_name(cls):
        raise NotImplementedError()

    @classmethod
    def _parse_header(cls, parsable):
        header_parser = ParserBinary(parsable)

        header_parser.parse_parsable('extension_name', SshCertExtensionName, 4)
        if header_parser['extension_name'] != cls.get_extension_name():
            raise InvalidType()

        return header_parser

    def _compose_header(self):
        header_composer = ComposerBinary()

        header_composer.compose_string(self.extension_name.value.code, 'ascii', 4)

        return header_composer


class SshCertExtensionNoData(SshCertExtensionParsed):
    @classmethod
    @abc.abstractmethod
    def get_extension_name(cls):
        raise NotImplementedError()

    @classmethod
    def _parse_header(cls, parsable):
        header_parser = super(SshCertExtensionNoData, cls)._parse_header(parsable)

        header_parser.parse_numeric('extension_length', 4)

        return header_parser

    def _compose_header(self):
        header_composer = super()._compose_header()

        header_composer.compose_numeric(0, 4)

        return header_composer

    @classmethod
    def _parse(cls, parsable):
        header_parser = cls._parse_header(parsable)

        return cls(), header_parser.parsed_length

    def compose(self):
        header_composer = self._compose_header()

        return header_composer.composed_bytes


class SshCertExtensionNoPrecenseRequired(SshCertExtensionNoData):
    @classmethod
    def get_extension_name(cls):
        return SshCertExtensionName.NO_PRESENCE_REQUIRED


class SshCertExtensionPermitX11Forwarding(SshCertExtensionNoData):
    @classmethod
    def get_extension_name(cls):
        return SshCertExtensionName.PERMIT_X11_FORWARDING


class SshCertExtensionPermitAgentForwarding(SshCertExtensionNoData):
    @classmethod
    def get_extension_name(cls):
        return SshCertExtensionName.PERMIT_AGENT_FORWARDING


class SshCertExtensionPermitPortForwarding(SshCertExtensionNoData):
    @classmethod
    def get_extension_name(cls):
        return SshCertExtensionName.PERMIT_PORT_FORWARDING


class SshCertExtensionPermitPTY(SshCertExtensionNoData):
    @classmethod
    def get_extension_name(cls):
        return SshCertExtensionName.PERMIT_PTY


class SshCertExtensionPermitUserRC(SshCertExtensionNoData):
    @classmethod
    def get_extension_name(cls):
        return SshCertExtensionName.PERMIT_USER_RC


@attr.s
class SshCertExtensionForceCommand(SshCertExtensionParsed):
    command = attr.ib(validator=attr.validators.instance_of(str))

    @classmethod
    def get_extension_name(cls):
        return SshCertExtensionName.FORCE_COMMAND

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_header(parsable)

        parser.parse_string('command', 4, 'ascii')

        return cls(parser['command']), parser.parsed_length

    def compose(self):
        body_composer = ComposerBinary()
        body_composer.compose_string(self.command, 'ascii', 4)

        header_composer = self._compose_header()

        return header_composer.composed_bytes + body_composer.composed_bytes


class VectorParamNetorkAddress(VectorParamString):
    def get_item_size(self, item):
        return len(str(item))


class NetworkVector(VectorString):
    @classmethod
    def get_param(cls):
        return VectorParamNetorkAddress(
            min_byte_num=0,
            max_byte_num=2 ** 32 - 1,
            separator=',',
            item_class=ipaddress.ip_network,
            fallback_class=None,
        )

    def compose(self):
        composer = ComposerBinary()

        address_composer = ComposerText()
        address_composer.compose_string_array(self._items)

        composer.compose_bytes(address_composer.composed, 4)

        return composer.composed


@attr.s
class SshCertExtensionSourceAddress(SshCertExtensionParsed):
    addresses = attr.ib(
        converter=NetworkVector,
        validator=attr.validators.instance_of(NetworkVector)
    )

    @classmethod
    def get_extension_name(cls):
        return SshCertExtensionName.SOURCE_ADDRESS

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_header(parsable)

        parser.parse_parsable('addresses', NetworkVector)

        return cls(parser['addresses']), parser.parsed_length

    def compose(self):
        composer = self._compose_header()

        composer.compose_parsable(self.addresses)

        return composer.composed_bytes


class SshCertConstraintVariant(VariantParsable):
    _VARIANTS = collections.OrderedDict([
        (SshCertExtensionName.NO_PRESENCE_REQUIRED, (SshCertExtensionNoPrecenseRequired, )),
        (SshCertExtensionName.PERMIT_X11_FORWARDING, (SshCertExtensionPermitX11Forwarding, )),
        (SshCertExtensionName.PERMIT_AGENT_FORWARDING, (SshCertExtensionPermitAgentForwarding, )),
        (SshCertExtensionName.PERMIT_PORT_FORWARDING, (SshCertExtensionPermitPortForwarding, )),
        (SshCertExtensionName.PERMIT_PTY, (SshCertExtensionPermitPTY, )),
        (SshCertExtensionName.PERMIT_USER_RC, (SshCertExtensionPermitUserRC, )),
        (SshCertExtensionName.FORCE_COMMAND, (SshCertExtensionForceCommand, )),
        (SshCertExtensionName.SOURCE_ADDRESS, (SshCertExtensionSourceAddress, )),
    ])

    @classmethod
    @abc.abstractmethod
    def _get_variants(cls):
        raise NotImplementedError()


class SshCertificateBase():
    @classmethod
    @abc.abstractmethod
    def _parse_host_key_algorithm(cls, parsable):
        raise NotImplementedError()

    @abc.abstractmethod
    def _compose_host_key_algorithm(self):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def _parse_host_key(cls, parser):
        raise NotImplementedError()

    @abc.abstractmethod
    def _compose_host_key_params(self, composer):
        raise NotImplementedError()

    def host_key_asdict(self):
        return attr.asdict(self, recurse=False, dict_factory=OrderedDict)


@attr.s
class SshHostCertificateV00Base(ParsableBase, SshCertificateBase):  # pylint: disable=too-many-instance-attributes
    certificate_type = attr.ib(validator=attr.validators.instance_of(SshCertType))
    key_id = attr.ib(
        validator=attr.validators.instance_of(str),
        metadata={'human_readable_name': 'Key ID'},
    )
    valid_principals = attr.ib(
        converter=SshCertValidPrincipals,
        validator=attr.validators.instance_of(SshCertValidPrincipals)
    )
    valid_after = attr.ib(validator=attr.validators.instance_of(datetime.datetime))
    valid_before = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(datetime.datetime)))
    constraints = attr.ib(
        converter=SshCertConstraintVector,
        validator=attr.validators.instance_of(SshCertConstraintVector)
    )
    nonce = attr.ib(validator=attr.validators.instance_of((bytes, bytearray)))
    reserved = attr.ib(validator=attr.validators.instance_of((bytes, bytearray)))
    signature_key = attr.ib(validator=attr.validators.instance_of(SshPublicKeyBase))
    signature = attr.ib(validator=attr.validators.instance_of(SshCertSignature))

    @classmethod
    @abc.abstractmethod
    def _parse_host_key_algorithm(cls, parsable):
        raise NotImplementedError()

    @abc.abstractmethod
    def _compose_host_key_algorithm(self):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def _parse_host_key(cls, parser):
        raise NotImplementedError()

    @abc.abstractmethod
    def _compose_host_key_params(self, composer):
        raise NotImplementedError()

    @classmethod
    def _parse_host_cert_params(cls, parser):
        parser.parse_parsable('certificate_type', SshCertTypeFactory)
        parser.parse_string('key_id', 4, 'ascii')
        parser.parse_parsable('valid_principals', SshCertValidPrincipals)

        parser.parse_timestamp('valid_after')
        parser.parse_timestamp('valid_before')

        parser.parse_parsable('constraints', SshCertConstraintVector)
        parser.parse_bytes('nonce', 4)
        parser.parse_bytes('reserved', 4)
        parser.parse_parsable('signature_key', SshHostPublicKeyVariant, 4)
        parser.parse_parsable('signature', SshCertSignature, 4)

    def _compose_host_cert_params(self, composer):
        composer.compose_parsable(self.certificate_type)
        composer.compose_string(self.key_id, 'ascii', 4)
        composer.compose_parsable(self.valid_principals)

        composer.compose_timestamp(self.valid_after)
        composer.compose_timestamp(self.valid_before)

        composer.compose_parsable(self.constraints)
        composer.compose_bytes(self.nonce, 4)
        composer.compose_bytes(self.reserved, 4)
        composer.compose_parsable(self.signature_key, 4)
        composer.compose_parsable(self.signature, 4)

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_host_key_algorithm(parsable)

        public_key = cls._parse_host_key(parser)

        cls._parse_host_cert_params(parser)

        return cls(public_key=public_key, **parser), parser.parsed_length

    def compose(self):
        composer = self._compose_host_key_algorithm()

        self._compose_host_key_params(composer)
        self._compose_host_cert_params(composer)

        return composer.composed


class SshHostCertificateBase():
    def _asdict(self):
        key_dict = OrderedDict([])

        key_dict.update(SshPublicKeyBase.host_key_asdict(self))
        key_dict.update(SshCertificateBase.host_key_asdict(self))

        return key_dict


@attr.s
class SshHostCertificateV00DSSBase(SshHostCertificateBase, SshHostKeyDSSBase, SshHostCertificateV00Base):
    @property
    def key_bytes(self):
        return self.compose()


class SshHostCertificateV00DSS(SshHostCertificateV00DSSBase, SshHostCertificateV00Base):
    @classmethod
    def get_host_key_algorithms(cls):
        return [SshHostKeyAlgorithm.SSH_DSS_CERT_V00_OPENSSH_COM, ]


@attr.s
class SshHostCertificateV00RSABase(SshHostCertificateBase, SshHostKeyRSABase, SshHostCertificateV00Base):
    @property
    def key_bytes(self):
        return self.compose()


@attr.s
class SshHostCertificateV00RSA(SshHostCertificateV00RSABase, SshHostCertificateV00Base):
    @classmethod
    def get_host_key_algorithms(cls):
        return [SshHostKeyAlgorithm.SSH_RSA_CERT_V00_OPENSSH_COM, ]


class SshCertExtensionVariant(SshCertConstraintVariant):
    @classmethod
    def _get_variants(cls):
        return collections.OrderedDict([
            (constraint_name, constraint_classes)
            for constraint_name, constraint_classes in cls._VARIANTS.items()
            if constraint_name.value.critical is False
        ])


class SshCertExtensionVector(VectorParsableDerived):
    @classmethod
    def get_param(cls):
        return VectorParamParsable(
            item_class=SshCertExtensionVariant,
            fallback_class=SshCertExtensionUnparsed,
            min_byte_num=0, max_byte_num=2 ** 32 - 1
        )


class SshCertCriticalOptionVariant(SshCertConstraintVariant):
    @classmethod
    def _get_variants(cls):
        return collections.OrderedDict([
            (constraint_name, constraint_classes)
            for constraint_name, constraint_classes in cls._VARIANTS.items()
            if constraint_name.value.critical is True
        ])


class SshCertCriticalOptionVector(VectorParsableDerived):
    @classmethod
    def get_param(cls):
        return VectorParamParsable(
            item_class=SshCertCriticalOptionVariant,
            fallback_class=SshCertExtensionUnparsed,
            min_byte_num=0, max_byte_num=2 ** 32 - 1
        )


@attr.s
class SshHostCertificateV01Base(ParsableBase, SshCertificateBase):  # pylint: disable=too-many-instance-attributes
    nonce = attr.ib(validator=attr.validators.instance_of((bytes, bytearray)))
    serial = attr.ib(validator=attr.validators.instance_of(int))
    certificate_type = attr.ib(validator=attr.validators.instance_of(SshCertType))
    key_id = attr.ib(validator=attr.validators.instance_of(str))
    valid_principals = attr.ib(
        converter=SshCertValidPrincipals,
        validator=attr.validators.instance_of(SshCertValidPrincipals)
    )
    valid_after = attr.ib(validator=attr.validators.instance_of(datetime.datetime))
    valid_before = attr.ib(validator=attr.validators.optional(attr.validators.instance_of(datetime.datetime)))
    critical_options = attr.ib(
        converter=SshCertCriticalOptionVector,
        validator=attr.validators.instance_of(SshCertCriticalOptionVector)
    )
    extensions = attr.ib(
        converter=SshCertExtensionVector,
        validator=attr.validators.instance_of(SshCertExtensionVector)
    )
    reserved = attr.ib(validator=attr.validators.instance_of((bytes, bytearray)))
    signature_key = attr.ib(validator=attr.validators.instance_of(SshPublicKeyBase))
    signature = attr.ib(
        validator=attr.validators.instance_of(SshCertSignature),
        metadata={'human_friendly': False},
    )

    @classmethod
    @abc.abstractmethod
    def _parse_host_key_algorithm(cls, parsable):
        raise NotImplementedError()

    @abc.abstractmethod
    def _compose_host_key_algorithm(self):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def _parse_host_key(cls, parser):
        raise NotImplementedError()

    @abc.abstractmethod
    def _compose_host_key_params(self, composer):
        raise NotImplementedError()

    @classmethod
    def _parse_host_cert_params(cls, parser):
        parser.parse_numeric('serial', 8)
        parser.parse_parsable('certificate_type', SshCertTypeFactory)
        parser.parse_string('key_id', 4, 'ascii')
        parser.parse_parsable('valid_principals', SshCertValidPrincipals)

        parser.parse_timestamp('valid_after')
        parser.parse_timestamp('valid_before')

        parser.parse_parsable('critical_options', SshCertCriticalOptionVector)
        parser.parse_parsable('extensions', SshCertExtensionVector)
        parser.parse_bytes('reserved', 4)
        parser.parse_parsable('signature_key', SshHostPublicKeyVariant, 4)
        parser.parse_parsable('signature', SshCertSignature, 4)

    def _compose_host_cert_params(self, composer):
        composer.compose_numeric(self.serial, 8)
        composer.compose_parsable(self.certificate_type)
        composer.compose_string(self.key_id, 'ascii', 4)
        composer.compose_parsable(self.valid_principals)

        composer.compose_timestamp(self.valid_after)
        composer.compose_timestamp(self.valid_before)

        composer.compose_parsable(self.critical_options)
        composer.compose_parsable(self.extensions)
        composer.compose_bytes(self.reserved, 4)
        composer.compose_parsable(self.signature_key, 4)
        composer.compose_parsable(self.signature, 4)

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_host_key_algorithm(parsable)

        parser.parse_bytes('nonce', 4)

        public_key = cls._parse_host_key(parser)

        cls._parse_host_cert_params(parser)

        return cls(public_key=public_key, **parser), parser.parsed_length

    def compose(self):
        composer = self._compose_host_key_algorithm()

        composer.compose_bytes(self.nonce, 4)

        self._compose_host_key_params(composer)
        self._compose_host_cert_params(composer)

        return composer.composed


@attr.s
class SshHostCertificateV01DSSBase(SshHostCertificateBase, SshHostKeyDSSBase, SshHostCertificateV01Base):
    @property
    def key_bytes(self):
        return self.compose()


class SshHostCertificateV01DSS(SshHostCertificateV01DSSBase, SshHostCertificateV01Base):
    @classmethod
    def get_host_key_algorithms(cls):
        return [SshHostKeyAlgorithm.SSH_DSS_CERT_V01_OPENSSH_COM, ]


@attr.s
class SshHostCertificateV01RSABase(SshHostCertificateBase, SshHostKeyRSABase, SshHostCertificateV01Base):
    @property
    def key_bytes(self):
        return self.compose()


@attr.s
class SshHostCertificateV01RSA(SshHostCertificateV01RSABase, SshHostCertificateV01Base):
    @classmethod
    def get_host_key_algorithms(cls):
        return [SshHostKeyAlgorithm.SSH_RSA_CERT_V01_OPENSSH_COM, ]


@attr.s
class SshHostCertificateV01ECDSABase(SshHostCertificateBase, SshHostKeyECDSABase, SshHostCertificateV01Base):
    @property
    def key_bytes(self):
        return self.compose()


class SshHostCertificateV01ECDSA(SshHostCertificateV01ECDSABase, SshHostCertificateV01Base):
    @classmethod
    def get_host_key_algorithms(cls):
        return [
            SshHostKeyAlgorithm.ECDSA_SHA2_NISTP256_CERT_V01_OPENSSH_COM,
            SshHostKeyAlgorithm.ECDSA_SHA2_NISTP384_CERT_V01_OPENSSH_COM,
            SshHostKeyAlgorithm.ECDSA_SHA2_NISTP521_CERT_V01_OPENSSH_COM,
            SshHostKeyAlgorithm.ECDSA_SHA2_SECP256K1_OID_CERT_V01_OPENSSH_COM,
        ]


@attr.s
class SshHostCertificateV01EDDSABase(SshHostCertificateBase, SshHostKeyEDDSABase, SshHostCertificateV01Base):
    @property
    def key_bytes(self):
        return self.compose()


@attr.s
class SshHostCertificateV01EDDSA(SshHostCertificateV01EDDSABase, SshHostCertificateV01Base):
    @classmethod
    def get_host_key_algorithms(cls):
        return [SshHostKeyAlgorithm.SSH_ED25519_CERT_V01_OPENSSH_COM, ]


@attr.s
class SshX509Certificate(ParsableBase, SshHostKeyBase):
    _NOT_DEFINED_HOST_KEY_ALGORITHMS_BY_PUBLIC_KEY_TYPE = {
        Authentication.RSA: SshHostKeyAlgorithm.X509V3_SIGN_RSA,
        Authentication.DSS: SshHostKeyAlgorithm.X509V3_SIGN_DSS,
    }

    @classmethod
    def get_host_key_algorithms(cls):
        return filter(
            lambda host_key_algorithm: host_key_algorithm.value.key_type == SshHostKeyType.X509_CERTIFICATE,
            SshHostKeyAlgorithm
        )

    @property
    def key_bytes(self):
        return self.public_key.public_key.der

    @classmethod
    def _parse(cls, parsable):
        try:
            parser = cls._parse_host_key_algorithm(parsable)
        except (InvalidValue, NotEnoughData):
            parser = ParserBinary(parsable)
            host_key_algorithm = None
        else:
            host_key_algorithm = parser['host_key_algorithm']

        if host_key_algorithm is None:
            public_key_length = parser.unparsed_length
        else:
            parser.parse_numeric('public_key_length', 4)
            public_key_length = parser['public_key_length']

        try:
            parser.parse_raw('public_key', public_key_length, PublicKeyX509.from_der)
        except (NotEnoughData, InvalidValue) as e:
            raise InvalidValue(parser.unparsed, cls, 'public_key') from e

        public_key = parser['public_key']
        if host_key_algorithm is None:
            host_key_algorithm = cls._NOT_DEFINED_HOST_KEY_ALGORITHMS_BY_PUBLIC_KEY_TYPE.get(public_key.key_type, None)
            if host_key_algorithm is None:
                raise InvalidType()

        return cls(host_key_algorithm, public_key), parser.parsed_length

    def compose(self):
        public_key_bytes = self.public_key.der

        if self.host_key_algorithm in [SshHostKeyAlgorithm.X509V3_SIGN_RSA, SshHostKeyAlgorithm.X509V3_SIGN_DSS]:
            composer = ComposerBinary()
        else:
            composer = self._compose_host_key_algorithm()
            composer.compose_numeric(len(public_key_bytes), 4)

        composer.compose_raw(public_key_bytes)

        return composer.composed


@attr.s
class SshX509CertificateChain(ParsableBase, SshHostKeyBase):
    issuer_certificates = attr.ib(
        validator=attr.validators.deep_iterable(member_validator=attr.validators.instance_of(PublicKey))
    )
    ocsp_responses = attr.ib(
        validator=attr.validators.deep_iterable(member_validator=attr.validators.instance_of((bytes, bytearray)))
    )

    @classmethod
    def get_host_key_algorithms(cls):
        return filter(
            lambda host_key_algorithm: host_key_algorithm.value.key_type == SshHostKeyType.X509_CERTIFICATE_CHAIN,
            SshHostKeyAlgorithm
        )

    @property
    def key_bytes(self):
        return self.public_key.key_bytes

    @classmethod
    def _parse(cls, parsable):
        parser = cls._parse_host_key_algorithm(parsable)

        parser.parse_numeric('certificate_count', 4)
        certificates = []
        for _ in range(parser['certificate_count']):
            parser.parse_bytes('certificate', 4)
            certificates.append(PublicKeyX509.from_der(bytes(parser['certificate'])))

        parser.parse_numeric('ocsp_response_count', 4)
        ocsp_responses = []
        for _ in range(parser['ocsp_response_count']):
            parser.parse_bytes('ocsp_response', 4)
            ocsp_responses.append(parser['ocsp_response'])

        return cls(
            parser['host_key_algorithm'],
            certificates[0], certificates[1:],
            ocsp_responses,
        ), parser.parsed_length

    def compose(self):
        composer = self._compose_host_key_algorithm()

        composer.compose_numeric(len(self.issuer_certificates) + 1, 4)
        for certificate in [self.public_key] + self.issuer_certificates:
            composer.compose_bytes(certificate.der, 4)

        composer.compose_numeric(len(self.ocsp_responses), 4)
        for ocsp_response in self.ocsp_responses:
            composer.compose_bytes(ocsp_response, 4)

        return composer.composed

    def _asdict(self):
        return collections.OrderedDict([
            ('key_type', self.host_key_algorithm.value.key_type.value),
            ('certificate_chain', [self.public_key] + self.issuer_certificates),
        ])


class SshHostPublicKeyVariant(VariantParsable):
    _VARIANTS = collections.OrderedDict(itertools.chain.from_iterable([
        [
            (host_key_algorithm, (ssh_key_class, ))
            for host_key_algorithm in ssh_key_class.get_host_key_algorithms()
        ]
        for ssh_key_class in [
            SshHostKeyDSS,
            SshHostKeyECDSA,
            SshHostKeyEDDSA,
            SshHostKeyRSA,
            SshHostCertificateV00DSS,
            SshHostCertificateV00RSA,
            SshHostCertificateV01DSS,
            SshHostCertificateV01ECDSA,
            SshHostCertificateV01EDDSA,
            SshHostCertificateV01RSA,
            SshX509Certificate,
            SshX509CertificateChain,
        ]
    ]))

    @classmethod
    def _get_variants(cls):
        return cls._VARIANTS



# SPDX-License-Identifier: MPL-2.0
# -*- coding: utf-8 -*-

import typing

import attr


@attr.s
class InvalidDataLength(Exception):
    bytes_needed: typing.Optional[int] = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(int))
    )


@attr.s
class NotEnoughData(InvalidDataLength):
    def __str__(self):
        return f'not enough data received from target; missing_byte_count="{self.bytes_needed}"'


@attr.s
class TooMuchData(InvalidDataLength):
    def __str__(self):
        return f'too much data received from target; rest_byte_count="{self.bytes_needed}"'


class InvalidType(Exception):
    def __str__(self):
        return 'invalid type value received from target'
